#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  self file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http:#www.gurux.org
#
#  self code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from ..enums import ObjectType, DataType
from .enums.GainResolution import GainResolution
from .enums.Modulation import Modulation
from ..internal._GXDataInfo import _GXDataInfo
from .GXDLMSMacPosTable import GXDLMSMacPosTable
from .GXDLMSNeighbourTable import GXDLMSNeighbourTable
from ..internal._GXLocalizer import _GXLocalizer

# pylint: disable=too-many-instance-attributes, too-many-public-methods
class GXDLMSG3PlcMacSetup(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    https://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSG3PlcMacSetup
    """

    def __init__(self, ln="0.0.29.1.0.255", sn=0):
        GXDLMSObject.__init__(self, ObjectType.G3_PLC_MAC_SETUP, ln, sn)
        self.version = 3
        self.__keyTable = []
        self.__shortAddress = 0xFFFF
        self.__rcCoord = 0xFFFF
        self.__panId = 0xFFFF
        self.__frameCounter = 0
        self.__tmrTtl = 2
        self.__maxFrameRetries = 5
        self.__neighbourTable = []
        self.__neighbourTableEntryTtl = 255
        self.__highPriorityWindowSize = 7
        self.__cscmFairnessLimit = 25
        self.__beaconRandomizationWindowLength = 12
        self.__a = 8
        self.__k = 5
        self.__minCwAttempts = 10
        self.__cenelecLegacyMode = 1
        self.__fccLegacyMode = 1
        self.__maxBe = 8
        self.__maxCsmaBackoffs = 50
        self.__minBe = 3
        self.__toneMask = ""
        self.__macBroadcastMaxCwEnabled = False
        self.__macTransmitAtten = 0
        self.__macPosTable = []
        self.__macDuplicateDetectionTtl = 0

    @classmethod
    def __getNeighbourTables(cls, tables):
        bb = GXByteBuffer()
        bb.setUInt8(DataType.ARRAY)
        if tables is None:
            bb.setUInt8(0)
        else:
            _GXCommon.setObjectCount(len(tables), bb)
            for it in tables:
                bb.setUInt8(DataType.STRUCTURE)
                bb.setUInt8(11)
                _GXCommon.setData(None, bb, DataType.UINT16, it.shortAddress)
                _GXCommon.setData(None, bb, DataType.BOOLEAN, it.enabled)
                _GXCommon.setData(None, bb, DataType.BITSTRING, it.toneMap)
                _GXCommon.setData(None, bb, DataType.ENUM, it.modulation)
                _GXCommon.setData(None, bb, DataType.INT8, it.txGain)
                _GXCommon.setData(None, bb, DataType.ENUM, it.txRes)
                _GXCommon.setData(None, bb, DataType.BITSTRING, it.txCoeff)
                _GXCommon.setData(None, bb, DataType.UINT8, it.lqi)
                _GXCommon.setData(None, bb, DataType.INT8, it.phaseDifferential)
                _GXCommon.setData(None, bb, DataType.UINT8, it.tmrValidTime)
                _GXCommon.setData(None, bb, DataType.UINT8, it.neighbourValidTime)
        return bb.array()

    @classmethod
    def __getPosTables(cls, tables):
        bb = GXByteBuffer()
        bb.setUInt8(DataType.ARRAY)
        if tables is None:
            bb.setUInt8(0)
        else:
            _GXCommon.setObjectCount(len(tables), bb)
            for it in tables:
                bb.setUInt8(DataType.STRUCTURE)
                bb.setUInt8(3)
                _GXCommon.setData(None, bb, DataType.UINT16, it.shortAddress)
                _GXCommon.setData(None, bb, DataType.UINT8, it.lQI)
                _GXCommon.setData(None, bb, DataType.UINT8, it.validTime)
        return bb.array()

    def invoke(self, settings, e):
        list_ = []
        if e.index == 1:
            index = e.value
            for it in self.__neighbourTable:
                if it.shortAddress == index:
                    list_.append(it)
            return self.__getNeighbourTables(list_)
        if e.index == 2:
            index = e.value
            for it in self.__macPosTable:
                if it.shortAddress == index:
                    list_.append(it)
            return self.__getPosTables(list_)
        e.error = ErrorCode.READ_WRITE_DENIED
        return None

    def getAttributeIndexToRead(self, all_):
        attributes = []
        # LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        # MacShortAddress
        if all_ or self.canRead(2):
            attributes.append(2)
        # MacRcCoord
        if all_ or self.canRead(3):
            attributes.append(3)
        # MacPANId
        if all_ or self.canRead(4):
            attributes.append(4)
        # MackeyTable
        if all_ or self.canRead(5):
            attributes.append(5)
        # MacFrameCounter
        if all_ or self.canRead(6):
            attributes.append(6)
        # MacToneMask
        if all_ or self.canRead(7):
            attributes.append(7)
        # MacTmrTtl
        if all_ or self.canRead(8):
            attributes.append(8)
        # MacMaxFrameRetries
        if all_ or self.canRead(9):
            attributes.append(9)
        # MacneighbourTableEntryTtl
        if all_ or self.canRead(10):
            attributes.append(10)
        # MacNeighbourTable
        if all_ or self.canRead(11):
            attributes.append(11)
        # MachighPriorityWindowSize
        if all_ or self.canRead(12):
            attributes.append(12)
        # MacCscmFairnessLimit
        if all_ or self.canRead(13):
            attributes.append(13)
        # MacBeaconRandomizationWindowLength
        if all_ or self.canRead(14):
            attributes.append(14)
        # MacA
        if all_ or self.canRead(15):
            attributes.append(15)
        # MacK
        if all_ or self.canRead(16):
            attributes.append(16)
        # MacMinCwAttempts
        if all_ or self.canRead(17):
            attributes.append(17)
        # MacCenelecLegacyMode
        if all_ or self.canRead(18):
            attributes.append(18)
        # MacFCCLegacyMode
        if all_ or self.canRead(19):
            attributes.append(19)
        # MacMaxBe
        if all_ or self.canRead(20):
            attributes.append(20)
        # MacMaxCsmaBackoffs,
        if all_ or self.canRead(21):
            attributes.append(21)
        # MacMinBe
        if all_ or self.canRead(22):
            attributes.append(22)
        # MacBroadcastMaxCwEnabled
        if all_ or self.canRead(23):
            attributes.append(23)
        # MacTransmitAtten
        if all_ or self.canRead(24):
            attributes.append(24)
        # MacPosTable
        if all_ or self.canRead(25):
            attributes.append(25)
        # MacDuplicateDetectionTtl
        if self.version > 2:
            if all_ or self.canRead(26):
                attributes.append(26)
        return attributes

    @classmethod
    def __parseNeighbourTableEntry(cls, value):
        list_ = []
        if value:
            for tmp in value:
                it = GXDLMSNeighbourTable()
                it.shortAddress = tmp[0]
                it.enabled = tmp[1]
                it.toneMap = str(tmp[2])
                it.modulation = Modulation(tmp[3])
                it.txGain = tmp[4]
                it.txRes = GainResolution(tmp[5])
                it.txCoeff = str(tmp[6])
                it.lqi = tmp[7]
                it.phaseDifferential = tmp[8]
                it.tmrValidTime = tmp[9]
                it.neighbourValidTime = tmp[10]
                list_.append(it)
        return list_

    @classmethod
    def __parsePosTableEntry(cls, value):
        list_ = []
        if value:
            for arr in value:
                it = GXDLMSMacPosTable()
                it.shortAddress = arr[0]
                it.lQI = arr[1]
                it.validTime = arr[2]
                list_.append(it)
        return list_

    def getAttributeCount(self):
        if self.version == 3:
            return 26
        if self.version == 2:
            return 25
        return 22

    def getMethodCount(self):
        if self.version == 3:
            return 2
        return 1

    def getValue(self, settings, e):
        if e.index == 1:
            ret = _GXCommon.logicalNameToBytes(self.logicalName)
        elif e.index == 2:
            ret = self.__shortAddress
        elif e.index == 3:
            ret = self.__rcCoord
        elif e.index == 4:
            ret = self.__panId
        elif e.index == 5:
            bb = GXByteBuffer()
            bb.setUInt8(DataType.ARRAY)
            if self.__keyTable is None:
                bb.setUInt8(0)
            else:
                _GXCommon.setObjectCount(len(self.__keyTable), bb)
                for k, v in self.__keyTable:
                    bb.setUInt8(DataType.STRUCTURE)
                    bb.setUInt8(2)
                    _GXCommon.setData(settings, bb, DataType.UINT8, k)
                    _GXCommon.setData(settings, bb, DataType.OCTET_STRING, v)
            ret = bb.array()
        elif e.index == 6:
            ret = self.__frameCounter
        elif e.index == 7:
            ret = self.__toneMask
        elif e.index == 8:
            ret = self.__tmrTtl
        elif e.index == 9:
            ret = self.__maxFrameRetries
        elif e.index == 10:
            ret = self.__neighbourTableEntryTtl
        elif e.index == 11:
            ret = self.__getNeighbourTables(self.__neighbourTable)
        elif e.index == 12:
            ret = self.__highPriorityWindowSize
        elif e.index == 13:
            ret = self.__cscmFairnessLimit
        elif e.index == 14:
            ret = self.__beaconRandomizationWindowLength
        elif e.index == 15:
            ret = self.__a
        elif e.index == 16:
            ret = self.__k
        elif e.index == 17:
            ret = self.__minCwAttempts
        elif e.index == 18:
            ret = self.__cenelecLegacyMode
        elif e.index == 19:
            ret = self.__fccLegacyMode
        elif e.index == 20:
            ret = self.__maxBe
        elif e.index == 21:
            ret = self.__maxCsmaBackoffs
        elif e.index == 22:
            ret = self.__minBe
        elif e.index == 23:
            ret = self.__macBroadcastMaxCwEnabled
        elif e.index == 24:
            ret = self.__macTransmitAtten
        elif e.index == 25:
            ret = self.__getPosTables(self.__macPosTable)
        elif e.index == 26:
            ret = self.__macDuplicateDetectionTtl
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
            ret = None
        return ret

    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.__shortAddress = e.value
        elif e.index == 3:
            self.__rcCoord = e.value
        elif e.index == 4:
            self.__panId = e.value
        elif e.index == 5:
            self.__keyTable.clear()
            if e.value:
                for arr in e.value:
                    self.__keyTable.append((arr[0], arr[1]))
        elif e.index == 6:
            self.frameCounter = e.value
        elif e.index == 7:
            self.toneMask = e.value
        elif e.index == 8:
            self.tmrTtl = e.value
        elif e.index == 9:
            self.maxFrameRetries = e.value
        elif e.index == 10:
            self.neighbourTableEntryTtl = e.value
        elif e.index == 11:
            self.__neighbourTable = self.__parseNeighbourTableEntry(e.value)
        elif e.index == 12:
            self.highPriorityWindowSize = e.value
        elif e.index == 13:
            self.cscmFairnessLimit = e.value
        elif e.index == 14:
            self.beaconRandomizationWindowLength = e.value
        elif e.index == 15:
            self.a = e.value
        elif e.index == 16:
            self.k = e.value
        elif e.index == 17:
            self.minCwAttempts = e.value
        elif e.index == 18:
            self.cenelecLegacyMode = e.value
        elif e.index == 19:
            self.fccLegacyMode = e.value
        elif e.index == 20:
            self.maxBe = e.value
        elif e.index == 21:
            self.maxCsmaBackoffs = e.value
        elif e.index == 22:
            self.minBe = e.value
        elif e.index == 23:
            self.macBroadcastMaxCwEnabled = e.value
        elif e.index == 24:
            self.macTransmitAtten = e.value
        elif e.index == 25:
            self.__macPosTable = self.__parsePosTableEntry(e.value)
        elif e.index == 26:
            self.macDuplicateDetectionTtl = e.value
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    @property
    def shortAddress(self):
        """
        The 16-bit address the device is using to communicate through the PAN.
        """
        return self.__shortAddress

    @shortAddress.setter
    def shortAddress(self, value):
        self.__shortAddress = value

    @property
    def rcCoord(self):
        """
        Route cost to coordinator.
        """
        return self.__rcCoord

    @rcCoord.setter
    def rcCoord(self, value):
        self.__rcCoord = value

    @property
    def pANId(self):
        """
        The 16-bit identifier of the PAN through which the device is operating.
        """
        return self.__panId

    @pANId.setter
    def pANId(self, value):
        self.__panId = value

    @property
    def keyTable(self):
        """
        self attribute holds GMK keys required for MAC layer ciphering.
        """
        return self.__keyTable

    @keyTable.setter
    def keyTable(self, value):
        self.__keyTable = value

    @property
    def frameCounter(self):
        """
        The outgoing frame counter for self device, used when ciphering frames at MAC layer.
        """
        return self.__frameCounter

    @frameCounter.setter
    def frameCounter(self, value):
        self.__frameCounter = value

    @property
    def toneMask(self):
        """
        Defines the tone mask to use during symbol formation.
        """
        return self.__toneMask

    @toneMask.setter
    def toneMask(self, value):
        self.__toneMask = value

    @property
    def tmrTtl(self):
        """
        Maximum time to live of tone map parameters entry in the neighbour table in minutes.
        """
        return self.__tmrTtl

    @tmrTtl.setter
    def tmrTtl(self, value):
        self.__tmrTtl = value

    @property
    def maxFrameRetries(self):
        """
        Maximum number of retransmissions.
        """
        return self.__maxFrameRetries

    @maxFrameRetries.setter
    def maxFrameRetries(self, value):
        self.__maxFrameRetries = value

    @property
    def neighbourTableEntryTtl(self):
        """
        Maximum time to live for an entry in the neighbour table in minutes
        """
        return self.__neighbourTableEntryTtl

    @neighbourTableEntryTtl.setter
    def neighbourTableEntryTtl(self, value):
        self.__neighbourTableEntryTtl = value

    @property
    def neighbourTable(self):
        """
        The neighbour table contains information about all the devices within the POS of the device
        """
        return self.__neighbourTable

    @neighbourTable.setter
    def neighbourTable(self, value):
        self.__neighbourTable = value

    @property
    def highPriorityWindowSize(self):
        """
        The high priority contention window size in number of slots.
        """
        return self.__highPriorityWindowSize

    @highPriorityWindowSize.setter
    def highPriorityWindowSize(self, value):
        self.__highPriorityWindowSize = value

    @property
    def cscmFairnessLimit(self):
        """
        Channel access fairness limit.
        """
        return self.__cscmFairnessLimit

    @cscmFairnessLimit.setter
    def cscmFairnessLimit(self, value):
        self.__cscmFairnessLimit = value

    @property
    def beaconRandomizationWindowLength(self):
        """
        Duration time in seconds for the beacon randomization.
        """
        return self.__beaconRandomizationWindowLength

    @beaconRandomizationWindowLength.setter
    def beaconRandomizationWindowLength(self, value):
        self.__beaconRandomizationWindowLength = value

    @property
    def a(self):
        """
        self parameter controls the adaptive CW linear decrease.
        """
        return self.__a

    @a.setter
    def a(self, value):
        self.__a = value

    @property
    def k(self):
        """
        Rate adaptation factor for channel access fairness limit.
        """
        return self.__k

    @k.setter
    def k(self, value):
        self.__k = value

    @property
    def minCwAttempts(self):
        """
        Number of consecutive attempts while using minimum CW.
        """
        return self.__minCwAttempts

    @minCwAttempts.setter
    def minCwAttempts(self, value):
        self.__minCwAttempts = value

    @property
    def cenelecLegacyMode(self):
        """
        self read only attribute indicates the capability of the node.
        """
        return self.__cenelecLegacyMode

    @cenelecLegacyMode.setter
    def cenelecLegacyMode(self, value):
        self.__cenelecLegacyMode = value

    @property
    def fccLegacyMode(self):
        """
        self read only attribute indicates the capability of the node.
        """
        return self.__fccLegacyMode

    @fccLegacyMode.setter
    def fccLegacyMode(self, value):
        self.__fccLegacyMode = value

    @property
    def maxBe(self):
        """
        Maximum value of backoff exponent.
        """
        return self.__maxBe

    @maxBe.setter
    def maxBe(self, value):
        self.__maxBe = value

    @property
    def maxCsmaBackoffs(self):
        """
        Maximum number of backoff attempts.
        """
        return self.__maxCsmaBackoffs

    @maxCsmaBackoffs.setter
    def maxCsmaBackoffs(self, value):
        self.__maxCsmaBackoffs = value

    @property
    def minBe(self):
        """
        Minimum value of backoff exponent.
        """
        return self.__minBe

    @minBe.setter
    def minBe(self, value):
        self.__minBe = value

    @property
    def macBroadcastMaxCwEnabled(self):
        """
        If True, MAC uses maximum contention window.
        """
        return self.__macBroadcastMaxCwEnabled

    @macBroadcastMaxCwEnabled.setter
    def macBroadcastMaxCwEnabled(self, value):
        self.__macBroadcastMaxCwEnabled = value

    @property
    def macTransmitAtten(self):
        """
        Attenuation of the output level in dB.
        """
        return self.__macTransmitAtten

    @macTransmitAtten.setter
    def macTransmitAtten(self, value):
        self.__macTransmitAtten = value

    @property
    def macPosTable(self):
        """
               The neighbour table contains some information

        about all the devices within the POS of the device.
        """
        return self.__macPosTable

    @macPosTable.setter
    def macPosTable(self, value):
        self.__macPosTable = value

    @property
    def macDuplicateDetectionTtl(self):
        """
        Duplicate frame detection time in seconds.
        """
        return self.__macDuplicateDetectionTtl

    @macDuplicateDetectionTtl.setter
    def macDuplicateDetectionTtl(self, value):
        self.__macDuplicateDetectionTtl = value

    def getValues(self):
        """
        Returns attributes as an array.

            Returns:
                Collection of COSEM object values.
        """
        return (
            self.logicalName,
            self.__shortAddress,
            self.__rcCoord,
            self.__panId,
            self.__keyTable,
            self.__frameCounter,
            self.__toneMask,
            self.__tmrTtl,
            self.__maxFrameRetries,
            self.__neighbourTableEntryTtl,
            self.__neighbourTable,
            self.__highPriorityWindowSize,
            self.__cscmFairnessLimit,
            self.__beaconRandomizationWindowLength,
            self.__a,
            self.__k,
            self.__minCwAttempts,
            self.__cenelecLegacyMode,
            self.__fccLegacyMode,
            self.__maxBe,
            self.__maxCsmaBackoffs,
            self.__minBe,
            self.__macBroadcastMaxCwEnabled,
            self.__macTransmitAtten,
            self.__macPosTable,
            self.__macDuplicateDetectionTtl,
        )

    def getNeighbourTableEntry(self, client, address):
        """
        Retrieves the MAC neighbour table.

            Parameters:
                client: DLMS client.
                address: MAC short address

            Returns:
                Generated bytes.
        """
        return client.method(self, 1, address)

    def parseNeighbourTableEntry(self, reply):
        """
        Parse neighbour table entry.

            Parameters:
                reply: Received reply

            Returns:
        """
        info = _GXDataInfo()
        value = _GXCommon.getData(None, reply, info)
        return self.__parseNeighbourTableEntry(value)

    def getPosTableEntry(self, client, address):
        """
        Retrieves the mac POS table.

            Parameters:
                client: DLMS client.
                address: MAC short address

            Returns:
                Generated bytes.
        """
        return client.method(self, 2, address)

    def parsePosTableEntry(self, reply):
        """
        Parse MAC POS tables.

            Parameters:
                reply: Received reply

            Returns:
        """
        info = _GXDataInfo()
        value = _GXCommon.getData(None, reply, info)
        return self.__parsePosTableEntry(value)

    def getNames(self):
        return (_GXLocalizer.gettext("Logical name"),\
            _GXLocalizer.gettext("MAC short address"),\
            _GXLocalizer.gettext("MAC rc coord"),\
            _GXLocalizer.gettext("MAC panid"),\
            _GXLocalizer.gettext("MAC key table"),\
            _GXLocalizer.gettext("MAC frame counter"),\
            _GXLocalizer.gettext("MAC tone mask"),\
            _GXLocalizer.gettext("MAC tmr ttl"),\
            _GXLocalizer.gettext("MAC max frame retries"),\
            _GXLocalizer.gettext("MAC neighbour table entry ttl"),\
            _GXLocalizer.gettext("MAC neighbour table"),\
            _GXLocalizer.gettext("MAC high priority window size"),\
            _GXLocalizer.gettext("MAC CSCM fairness limit"),\
            _GXLocalizer.gettext("MAC beacon randomization window length"),\
            _GXLocalizer.gettext("MAC a"),\
            _GXLocalizer.gettext("MAC k"),\
            _GXLocalizer.gettext("MAC min cw attempts"),\
            _GXLocalizer.gettext("MAC cenelec legacy mode"),\
            _GXLocalizer.gettext("MAC fcclegacy mode"),\
            _GXLocalizer.gettext("MAC max be"),\
            _GXLocalizer.gettext("MAC max csma backoffs"),\
            _GXLocalizer.gettext("MAC min be"),\
            _GXLocalizer.gettext("MAC broadcast max cw enabled"),\
            _GXLocalizer.gettext("MAC transmit atten"),\
            _GXLocalizer.gettext("MAC pos table"),\
            _GXLocalizer.gettext("MAC duplicate detection ttl"))

    def getMethodNames(self):
        return (_GXLocalizer.gettext("MAC get neighbour table entry"),\
            _GXLocalizer.gettext("MAC get pos tableentry"))

    def getDataType(self, index):
        if index == 1:
            ret = DataType.OCTET_STRING
        elif index == 2:
            ret = DataType.UINT16
        elif index == 3:
            ret = DataType.UINT16
        elif index == 4:
            ret = DataType.UINT16
        elif index == 5:
            ret = DataType.ARRAY
        elif index == 6:
            ret = DataType.UINT32
        elif index == 7:
            ret = DataType.BITSTRING
        elif index == 8:
            ret = DataType.UINT8
        elif index == 9:
            ret = DataType.UINT8
        elif index == 10:
            ret = DataType.UINT8
        elif index == 11:
            ret = DataType.ARRAY
        elif index == 12:
            ret = DataType.UINT8
        elif index == 13:
            ret = DataType.UINT8
        elif index == 14:
            ret = DataType.UINT8
        elif index == 15:
            ret = DataType.UINT8
        elif index == 16:
            ret = DataType.UINT8
        elif index == 17:
            ret = DataType.UINT8
        elif index == 18:
            ret = DataType.UINT8
        elif index == 19:
            ret = DataType.UINT8
        elif index == 20:
            ret = DataType.UINT8
        elif index == 21:
            ret = DataType.UINT8
        elif index == 22:
            ret = DataType.UINT8
        elif index == 23:
            ret = DataType.BOOLEAN
        elif index == 24:
            ret = DataType.UINT8
        elif index == 25:
            ret = DataType.ARRAY
        elif index == 26:
            ret = DataType.UINT8
        else:
            raise ValueError("GetDataType failed. Invalid attribute index.")
        return ret

    def __loadKeyTable(self, reader):
        self.__keyTable.clear()
        if reader.isStartElement("KeyTable", True):
            while reader.isStartElement("Item", True):
                k = reader.readElementContentAsInt("Key")
                d = GXByteBuffer.hexToBytes(reader.readElementContentAsString("Data"))
                self.__keyTable.append((k, d))
            reader.readEndElement("KeyTable")

    def __loadNeighbourTable(self, reader):
        self.__neighbourTable.clear()
        if reader.isStartElement("NeighbourTable", True):
            while reader.isStartElement("Item", True):
                it = GXDLMSNeighbourTable()
                it.shortAddress = reader.readElementContentAsInt("ShortAddress")
                it.enabled = reader.readElementContentAsInt("Enabled") != 0
                it.toneMap = reader.readElementContentAsString("ToneMap")
                it.modulation = Modulation(reader.readElementContentAsInt("Modulation"))
                it.txGain = reader.readElementContentAsInt("TxGain")
                it.txRes = GainResolution(reader.readElementContentAsInt("TxRes"))
                it.txCoeff = reader.readElementContentAsString("TxCoeff")
                it.lqi = reader.readElementContentAsInt("Lqi")
                it.phaseDifferential = reader.readElementContentAsInt(
                    "PhaseDifferential"
                )
                it.tmrValidTime = reader.readElementContentAsInt("TMRValidTime")
                it.NeighbourValidTime = reader.readElementContentAsInt(
                    "NeighbourValidTime"
                )
                self.__neighbourTable.Add(it)
            reader.readEndElement("NeighbourTable")

    def __loadMacPosTable(self, reader):
        self.__macPosTable.clear()
        if reader.isStartElement("MacPosTable", True):
            while reader.isStartElement("Item", True):
                it = GXDLMSMacPosTable()
                it.shortAddress = reader.readElementContentAsInt("ShortAddress")
                it.lqi = reader.readElementContentAsInt("LQI")
                it.validTime = reader.readElementContentAsInt("ValidTime")
                self.__macPosTable.Add(it)
            reader.readEndElement("MacPosTable")

    def load(self, reader):
        self.__shortAddress = reader.readElementContentAsInt("ShortAddress")
        self.__rcCoord = reader.readElementContentAsInt("RcCoord")
        self.__panId = reader.readElementContentAsInt("PANId")
        self.__loadKeyTable(reader)
        self.__frameCounter = reader.readElementContentAsInt("FrameCounter")
        self.__toneMask = reader.readElementContentAsString("ToneMask")
        self.__tmrTtl = reader.readElementContentAsInt("TmrTtl")
        self.__maxFrameRetries = reader.readElementContentAsInt("MaxFrameRetries")
        self.__neighbourTableEntryTtl = reader.readElementContentAsInt(
            "NeighbourTableEntryTtl"
        )
        self.__loadNeighbourTable(reader)
        self.__highPriorityWindowSize = reader.readElementContentAsInt(
            "HighPriorityWindowSize"
        )
        self.__cscmFairnessLimit = reader.readElementContentAsInt("CscmFairnessLimit")
        self.__beaconRandomizationWindowLength = reader.readElementContentAsInt(
            "BeaconRandomizationWindowLength"
        )
        self.__a = reader.readElementContentAsInt("A")
        self.__k = reader.readElementContentAsInt("K")
        self.__minCwAttempts = reader.readElementContentAsInt("MinCwAttempts")
        self.__cenelecLegacyMode = reader.readElementContentAsInt("CenelecLegacyMode")
        self.__fccLegacyMode = reader.readElementContentAsInt("FccLegacyMode")
        self.__maxBe = reader.readElementContentAsInt("MaxBe")
        self.__maxCsmaBackoffs = reader.readElementContentAsInt("MaxCsmaBackoffs")
        self.__minBe = reader.readElementContentAsInt("MinBe")
        self.__macBroadcastMaxCwEnabled = (
            reader.readElementContentAsInt("MacBroadcastMaxCwEnabled") != 0
        )
        self.__macTransmitAtten = reader.readElementContentAsInt("MacTransmitAtten")
        self.__loadMacPosTable(reader)
        self.__macDuplicateDetectionTtl = reader.readElementContentAsInt(
            "MacDuplicateDetectionTtl"
        )

    def __saveKeyTable(self, writer):
        writer.writeStartElement("KeyTable")
        if self.__keyTable:
            for k, v in self.__keyTable:
                writer.writeStartElement("Item")
                writer.writeElementString("Key", k)
                writer.writeElementString("Data", GXByteBuffer.hex(v))
                writer.writeEndElement()
        writer.writeEndElement()  # KeyTable

    def __saveNeighbourTable(self, writer):
        writer.writeStartElement("NeighbourTable")
        if self.__neighbourTable:
            for it in self.__neighbourTable:
                writer.writeStartElement("Item")
                writer.writeElementString("ShortAddress", it.shortAddress)
                writer.writeElementString("Enabled", it.enabled)
                writer.writeElementString("ToneMap", it.toneMap)
                writer.writeElementString("Modulation", it.modulation)
                writer.writeElementString("TxGain", it.txGain)
                writer.writeElementString("TxRes", it.txRes)
                writer.writeElementString("TxCoeff", it.txCoeff)
                writer.writeElementString("Lqi", it.Lqi)
                writer.writeElementString("PhaseDifferential", it.phaseDifferential)
                writer.writeElementString("TMRValidTime", it.tmrValidTime)
                writer.writeElementString("NeighbourValidTime", it.neighbourValidTime)
                writer.writeEndElement()
        writer.writeEndElement()  # NeighbourTable

    def __saveMacPosTable(self, writer):
        writer.writeStartElement("MacPosTable")
        if self.__macPosTable:
            for it in self.__macPosTable:
                writer.writeStartElement("Item")
                writer.writeElementString("ShortAddress", it.shortAddress)
                writer.writeElementString("LQI", it.lqi)
                writer.writeElementString("ValidTime", it.validTime)
                writer.writeEndElement()
        writer.writeEndElement()  # MacPosTable

    def save(self, writer):
        writer.writeElementString("ShortAddress", self.__shortAddress)
        writer.writeElementString("RcCoord", self.__rcCoord)
        writer.writeElementString("PANId", self.__panId)
        self.__saveKeyTable(writer)
        writer.writeElementString("FrameCounter", self.__frameCounter)
        writer.writeElementString("ToneMask", self.__toneMask)
        writer.writeElementString("TmrTtl", self.__tmrTtl)
        writer.writeElementString("MaxFrameRetries", self.__maxFrameRetries)
        writer.writeElementString(
            "NeighbourTableEntryTtl", self.__neighbourTableEntryTtl
        )
        self.__saveNeighbourTable(writer)
        writer.writeElementString(
            "HighPriorityWindowSize", self.__highPriorityWindowSize
        )
        writer.writeElementString("CscmFairnessLimit", self.__cscmFairnessLimit)
        writer.writeElementString(
            "BeaconRandomizationWindowLength", self.__beaconRandomizationWindowLength
        )
        writer.writeElementString("A", self.__a)
        writer.writeElementString("K", self.__k)
        writer.writeElementString("MinCwAttempts", self.__minCwAttempts)
        writer.writeElementString("CenelecLegacyMode", self.__cenelecLegacyMode)
        writer.writeElementString("FccLegacyMode", self.__fccLegacyMode)
        writer.writeElementString("MaxBe", self.__maxBe)
        writer.writeElementString("MaxCsmaBackoffs", self.__maxCsmaBackoffs)
        writer.writeElementString("MinBe", self.__minBe)
        writer.writeElementString(
            "MacBroadcastMaxCwEnabled", self.__macBroadcastMaxCwEnabled
        )
        writer.writeElementString("MacTransmitAtten", self.__macTransmitAtten)
        self.__saveMacPosTable(writer)
        writer.writeElementString(
            "MacDuplicateDetectionTtl", self.__macDuplicateDetectionTtl
        )



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http:#www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from gurux_dlms.asn.enums.PkcsObjectIdentifier import PkcsObjectIdentifier


class PkcsObjectIdentifierConverter:
    @classmethod
    def getString(cls, value):
        if value == PkcsObjectIdentifier.RSA_ENCRYPTION:
            ret = "1.2.840.113549.1.1.1"
        elif value == PkcsObjectIdentifier.MD2WITH_RSA_ENCRYPTION:
            ret = "1.2.840.113549.1.1.2"
        elif value == PkcsObjectIdentifier.MD4WITH_RSA_ENCRYPTION:
            ret = "1.2.840.113549.1.1.3"
        elif value == PkcsObjectIdentifier.MD5WITH_RSA_ENCRYPTION:
            ret = "1.2.840.113549.1.1.4"
        elif value == PkcsObjectIdentifier.SHA1WITH_RSA_ENCRYPTION:
            ret = "1.2.840.113549.1.1.5"
        elif value == PkcsObjectIdentifier.SRSA_OAEP_ENCRYPTION_SET:
            ret = "1.2.840.113549.1.1.6"
        elif value == PkcsObjectIdentifier.ID_RSAES_OAEP:
            ret = "1.2.840.113549.1.1.7"
        elif value == PkcsObjectIdentifier.ID_MGF1:
            ret = "1.2.840.113549.1.1.8"
        elif value == PkcsObjectIdentifier.ID_P_SPECIFIED:
            ret = "1.2.840.113549.1.1.9"
        elif value == PkcsObjectIdentifier.ID_RSASSA_PSS:
            ret = "1.2.840.113549.1.1.10"
        elif value == PkcsObjectIdentifier.SHA256WITH_RSA_ENCRYPTION:
            ret = "1.2.840.113549.1.1.11"
        elif value == PkcsObjectIdentifier.SHA384WITH_RSA_ENCRYPTION:
            ret = "1.2.840.113549.1.1.12"
        elif value == PkcsObjectIdentifier.SHA512WITH_RSA_ENCRYPTION:
            ret = "1.2.840.113549.1.1.13"
        elif value == PkcsObjectIdentifier.SHA224WITH_RSA_ENCRYPTION:
            ret = "1.2.840.113549.1.1.14"
        elif value == PkcsObjectIdentifier.DH_KEY_AGREE1MENT:
            ret = "1.2.840.113549.1.3.1"
        elif value == PkcsObjectIdentifier.PBE_WITH_MD2AND_DES_CBC:
            ret = "1.2.840.113549.1.5.1"
        elif value == PkcsObjectIdentifier.PBE_WITH_MD2AND_RC2CBC:
            ret = "1.2.840.113549.1.5.4"
        elif value == PkcsObjectIdentifier.PBE_WITH_MD5AND_DES_CBC:
            ret = "1.2.840.113549.1.5.3"
        elif value == PkcsObjectIdentifier.PBE_WITH_MD5AND_RC2CBC:
            ret = "1.2.840.113549.1.5.6"
        elif value == PkcsObjectIdentifier.PBE_WITH_SHA1AND_DES_CBC:
            ret = "1.2.840.113549.1.5.10"
        elif value == PkcsObjectIdentifier.PBE_WITH_SHA1AND_RC2CBC:
            ret = "1.2.840.113549.1.5.11"
        elif value == PkcsObjectIdentifier.ID_PBE_S2:
            ret = "1.2.840.113549.1.5.13"
        elif value == PkcsObjectIdentifier.ID_PBKDF2:
            ret = "1.2.840.113549.1.5.12"
        elif value == PkcsObjectIdentifier.DES_EDE3CBC:
            ret = "1.2.840.113549.3.7"
        elif value == PkcsObjectIdentifier.RC2CBC:
            ret = "1.2.840.113549.3.2"
        elif value == PkcsObjectIdentifier.MD2:
            ret = "1.2.840.113549.2.2"
        elif value == PkcsObjectIdentifier.MD4:
            ret = "1.2.840.113549.2.4"
        elif value == PkcsObjectIdentifier.MD5:
            ret = "1.2.840.113549.2.5"
        elif value == PkcsObjectIdentifier.ID_HMAC_WITH_SHA1:
            ret = "1.2.840.113549.2.7"
        elif value == PkcsObjectIdentifier.ID_HMAC_WITH_SHA224:
            ret = "1.2.840.113549.2.8"
        elif value == PkcsObjectIdentifier.ID_HMAC_WITH_SHA256:
            ret = "1.2.840.113549.2.9"
        elif value == PkcsObjectIdentifier.ID_HMAC_WITH_SHA384:
            ret = "1.2.840.113549.2.10"
        elif value == PkcsObjectIdentifier.ID_HMAC_WITH_SHA512:
            ret = "1.2.840.113549.2.11"
        elif value == PkcsObjectIdentifier.DATA:
            ret = "1.2.840.113549.1.7.1"
        elif value == PkcsObjectIdentifier.SIGNED_DATA:
            ret = "1.2.840.113549.1.7.2"
        elif value == PkcsObjectIdentifier.ENVELOPED_DATA:
            ret = "1.2.840.113549.1.7.3"
        elif value == PkcsObjectIdentifier.SIGNED_AND_ENVELOPED_DATA:
            ret = "1.2.840.113549.1.7.4"
        elif value == PkcsObjectIdentifier.DIGESTED_DATA:
            ret = "1.2.840.113549.1.7.5"
        elif value == PkcsObjectIdentifier.ENCRYPTED_DATA:
            ret = "1.2.840.113549.1.7.6"
        elif value == PkcsObjectIdentifier.PKCS9AT_EMAIL_ADDRESS:
            ret = "1.2.840.113549.1.9.1"
        elif value == PkcsObjectIdentifier.PKCS9AT_UNSTRUCTURED_NAME:
            ret = "1.2.840.113549.1.9.2"
        elif value == PkcsObjectIdentifier.PKCS9AT_CONTENT_TYPE:
            ret = "1.2.840.113549.1.9.3"
        elif value == PkcsObjectIdentifier.PKCS9AT_MESSAGE_DIGEST:
            ret = "1.2.840.113549.1.9.4"
        elif value == PkcsObjectIdentifier.PKCS9AT_SIGNING_TIME:
            ret = "1.2.840.113549.1.9.5"
        elif value == PkcsObjectIdentifier.PKCS9AT_COUNTER_SIGNATURE:
            ret = "1.2.840.113549.1.9.6"
        elif value == PkcsObjectIdentifier.PKCS9AT_CHALLENGE_PASSWORD:
            ret = "1.2.840.113549.1.9.7"
        elif value == PkcsObjectIdentifier.PKCS9AT_UNSTRUCTURED_ADDRESS:
            ret = "1.2.840.113549.1.9.8"
        elif value == PkcsObjectIdentifier.PKCS9AT_EXTENDED_CERTIFICATE_ATTRIBUTES:
            ret = "1.2.840.113549.1.9.9"
        elif value == PkcsObjectIdentifier.PKCS9AT_SIGNING_DESCRIPTION:
            ret = "1.2.840.113549.1.9.13"
        elif value == PkcsObjectIdentifier.PKCS9AT_EXTENSION_REQUEST:
            ret = "1.2.840.113549.1.9.14"
        elif value == PkcsObjectIdentifier.PKCS9AT_SMIME_CAPABILITIES:
            ret = "1.2.840.113549.1.9.15"
        elif value == PkcsObjectIdentifier.ID_SMIME:
            ret = "1.2.840.113549.1.9.16"
        elif value == PkcsObjectIdentifier.PKCS9AT_FRIENDLY_NAME:
            ret = "1.2.840.113549.1.9.20"
        elif value == PkcsObjectIdentifier.PKCS9AT_LOCAL_KEY_ID:
            ret = "1.2.840.113549.1.9.21"
        elif value == PkcsObjectIdentifier.X509CERTIFICATE:
            ret = "1.2.840.113549.1.9.22.1"
        elif value == PkcsObjectIdentifier.SDSI_CERTIFICATE:
            ret = "1.2.840.113549.1.9.22.2"
        elif value == PkcsObjectIdentifier.X509CRL:
            ret = "1.2.840.113549.1.9.23.1"
        elif value == PkcsObjectIdentifier.ID_ALG:
            ret = "1.2.840.113549.1.9.16.3"
        elif value == PkcsObjectIdentifier.ID_ALG_ESDH:
            ret = "1.2.840.113549.1.9.16.3.5"
        elif value == PkcsObjectIdentifier.ID_ALG_CMS3DES_WRAP:
            ret = "1.2.840.113549.1.9.16.3.6"
        elif value == PkcsObjectIdentifier.ID_ALG_CMS_RC2WRAP:
            ret = "1.2.840.113549.1.9.16.3.7"
        elif value == PkcsObjectIdentifier.ID_ALG_PWRI_KEK:
            ret = "1.2.840.113549.1.9.16.3.9"
        elif value == PkcsObjectIdentifier.ID_ALG_SSDH:
            ret = "1.2.840.113549.1.9.16.3.10"
        elif value == PkcsObjectIdentifier.ID_RSA_KEM:
            ret = "1.2.840.113549.1.9.16.3.14"
        elif value == PkcsObjectIdentifier.PREFER_SIGNED_DATA:
            ret = "1.2.840.113549.1.9.15.1"
        elif value == PkcsObjectIdentifier.CANNOT_DECRYPT_ANY:
            ret = "1.2.840.113549.1.9.15.2"
        elif value == PkcsObjectIdentifier.SMIME_CAPABILITIES_VERSIONS:
            ret = "1.2.840.113549.1.9.15.3"
        elif value == PkcsObjectIdentifier.ID_AA_RECEIPT_REQUEST:
            ret = "1.2.840.113549.1.9.16.2.1"
        elif value == PkcsObjectIdentifier.ID_CT_AUTH_DATA:
            ret = "1.2.840.113549.1.9.16.1.2"
        elif value == PkcsObjectIdentifier.ID_CT_TST_INFO:
            ret = "1.2.840.113549.1.9.16.1.4"
        elif value == PkcsObjectIdentifier.ID_CT_COMPRESSED_DATA:
            ret = "1.2.840.113549.1.9.16.1.9"
        elif value == PkcsObjectIdentifier.ID_CT_AUTH_ENVELOPED_DATA:
            ret = "1.2.840.113549.1.9.16.1.23"
        elif value == PkcsObjectIdentifier.ID_CT_TIMESTAMPED_DATA:
            ret = "1.2.840.113549.1.9.16.1.31"
        elif value == PkcsObjectIdentifier.ID_CTI_ETS_PROOF_OF_ORIGIN:
            ret = "1.2.840.113549.1.9.16.6.1"
        elif value == PkcsObjectIdentifier.ID_CTI_ETS_PROOF_OF_RECEIPT:
            ret = "1.2.840.113549.1.9.16.6.2"
        elif value == PkcsObjectIdentifier.ID_CTI_ETS_PROOF_OF_DELIVERY:
            ret = "1.2.840.113549.1.9.16.6.3"
        elif value == PkcsObjectIdentifier.ID_CTI_ETS_PROOF_OF_SENDER:
            ret = "1.2.840.113549.1.9.16.6.4"
        elif value == PkcsObjectIdentifier.ID_CTI_ETS_PROOF_OF_APPROVAL:
            ret = "1.2.840.113549.1.9.16.6.5"
        elif value == PkcsObjectIdentifier.ID_CTI_ETS_PROOF_OF_CREATION:
            ret = "1.2.840.113549.1.9.16.6.6"
        elif value == PkcsObjectIdentifier.ID_AA_CONTENT_HINT:
            ret = "1.2.840.113549.1.9.16.2.4"
        elif value == PkcsObjectIdentifier.ID_AA_MSG_SIG_DIGEST:
            ret = "1.2.840.113549.1.9.16.2.5"
        elif value == PkcsObjectIdentifier.ID_AA_CONTENT_REFERENCE:
            ret = "1.2.840.113549.1.9.16.2.10"
        elif value == PkcsObjectIdentifier.ID_AA_ENCRYP_KEY_PREF:
            ret = "1.2.840.113549.1.9.16.2.11"
        elif value == PkcsObjectIdentifier.ID_AA_SIGNING_CERTIFICATE:
            ret = "1.2.840.113549.1.9.16.2.12"
        elif value == PkcsObjectIdentifier.ID_AA_SIGNING_CERTIFICATE_V2:
            ret = "1.2.840.113549.1.9.16.2.47"
        elif value == PkcsObjectIdentifier.ID_AA_CONTENT_IDENTIFIER:
            ret = "1.2.840.113549.1.9.16.2.7"
        elif value == PkcsObjectIdentifier.ID_AA_SIGNATURE_TIME_STAMP_TOKEN:
            ret = "1.2.840.113549.1.9.16.2.14"
        elif value == PkcsObjectIdentifier.ID_AA_ETS_SIG_POLICY_ID:
            ret = "1.2.840.113549.1.9.16.2.15"
        elif value == PkcsObjectIdentifier.ID_AA_ETS_COMMITMENT_TYPE:
            ret = "1.2.840.113549.1.9.16.2.16"
        elif value == PkcsObjectIdentifier.ID_AA_ETS_SIGNER_LOCATION:
            ret = "1.2.840.113549.1.9.16.2.17"
        elif value == PkcsObjectIdentifier.ID_AA_ETS_SIGNER_ATTR:
            ret = "1.2.840.113549.1.9.16.2.18"
        elif value == PkcsObjectIdentifier.ID_AA_ETS_OTHER_SIG_CERT:
            ret = "1.2.840.113549.1.9.16.2.19"
        elif value == PkcsObjectIdentifier.ID_AA_ETS_CONTENT_TIMESTAMP:
            ret = "1.2.840.113549.1.9.16.2.20"
        elif value == PkcsObjectIdentifier.ID_AA_ETS_CERTIFICATE_REFS:
            ret = "1.2.840.113549.1.9.16.2.21"
        elif value == PkcsObjectIdentifier.ID_AA_ETS_REVOCATION_REFS:
            ret = "1.2.840.113549.1.9.16.2.22"
        elif value == PkcsObjectIdentifier.ID_AA_ETS_CERT_VALUES:
            ret = "1.2.840.113549.1.9.16.2.23"
        elif value == PkcsObjectIdentifier.ID_AA_ETS_REVOCATION_VALUES:
            ret = "1.2.840.113549.1.9.16.2.24"
        elif value == PkcsObjectIdentifier.ID_AA_ETS_ESC_TIME_STAMP:
            ret = "1.2.840.113549.1.9.16.2.25"
        elif value == PkcsObjectIdentifier.ID_AA_ETS_CERT_CRL_TIMESTAMP:
            ret = "1.2.840.113549.1.9.16.2.26"
        elif value == PkcsObjectIdentifier.ID_AA_ETS_ARCHIVE_TIMESTAMP:
            ret = "1.2.840.113549.1.9.16.2.27"
        elif value == PkcsObjectIdentifier.ID_SPQ_ETS_URI:
            ret = "1.2.840.113549.1.9.16.5.1"
        elif value == PkcsObjectIdentifier.ID_SPQ_ETS_U_NOTICE:
            ret = "1.2.840.113549.1.9.16.5.2"
        elif value == PkcsObjectIdentifier.KEY_BAG:
            ret = "1.2.840.113549.1.12.10.1.1"
        elif value == PkcsObjectIdentifier.PKCS8SHROUDED_KEY_BAG:
            ret = "1.2.840.113549.1.12.10.1.2"
        elif value == PkcsObjectIdentifier.CERT_BAG:
            ret = "1.2.840.113549.1.12.10.1.3"
        elif value == PkcsObjectIdentifier.CRL_BAG:
            ret = "1.2.840.113549.1.12.10.1.4"
        elif value == PkcsObjectIdentifier.SECRET_BAG:
            ret = "1.2.840.113549.1.12.10.1.5"
        elif value == PkcsObjectIdentifier.SAFE_CONTENTS_BAG:
            ret = "1.2.840.113549.1.12.10.1.6"
        elif value == PkcsObjectIdentifier.PBE_WITH_SHA_AND128BIT_RC4:
            ret = "1.2.840.113549.1.12.1.1"
        elif value == PkcsObjectIdentifier.PBE_WITH_SHA_AND40BIT_RC4:
            ret = "1.2.840.113549.1.12.1.2"
        elif value == PkcsObjectIdentifier.PBE_WITH_SHA_AND3KEY_TRIPLE_DES_CBC:
            ret = "1.2.840.113549.1.12.1.3"
        elif value == PkcsObjectIdentifier.PBE_WITH_SHA_AND2KEY_TRIPLE_DES_CBC:
            ret = "1.2.840.113549.1.12.1.4"
        elif value == PkcsObjectIdentifier.PBE_WITH_SHA_AND128BIT_RC2CBC:
            ret = "1.2.840.113549.1.12.1.5"
        elif value == PkcsObjectIdentifier.PBEWITH_SHA_AND40BIT_RC2CBC:
            ret = "1.2.840.113549.1.12.1.6"
        else:
            ret = PkcsObjectIdentifier.NONE
        return ret

    @classmethod
    def fromString(cls, value):
        if value == "1.2.840.113549.1.1.1":
            ret = PkcsObjectIdentifier.RSA_ENCRYPTION
        elif value == "1.2.840.113549.1.1.2":
            ret = PkcsObjectIdentifier.MD2WITH_RSA_ENCRYPTION
        elif value == "1.2.840.113549.1.1.3":
            ret = PkcsObjectIdentifier.MD4WITH_RSA_ENCRYPTION
        elif value == "1.2.840.113549.1.1.4":
            ret = PkcsObjectIdentifier.MD5WITH_RSA_ENCRYPTION
        elif value == "1.2.840.113549.1.1.5":
            ret = PkcsObjectIdentifier.SHA1WITH_RSA_ENCRYPTION
        elif value == "1.2.840.113549.1.1.6":
            ret = PkcsObjectIdentifier.SRSA_OAEP_ENCRYPTION_SET
        elif value == "1.2.840.113549.1.1.7":
            ret = PkcsObjectIdentifier.ID_RSAES_OAEP
        elif value == "1.2.840.113549.1.1.8":
            ret = PkcsObjectIdentifier.ID_MGF1
        elif value == "1.2.840.113549.1.1.9":
            ret = PkcsObjectIdentifier.ID_P_SPECIFIED
        elif value == "1.2.840.113549.1.1.10":
            ret = PkcsObjectIdentifier.ID_RSASSA_PSS
        elif value == "1.2.840.113549.1.1.11":
            ret = PkcsObjectIdentifier.SHA256WITH_RSA_ENCRYPTION
        elif value == "1.2.840.113549.1.1.12":
            ret = PkcsObjectIdentifier.SHA384WITH_RSA_ENCRYPTION
        elif value == "1.2.840.113549.1.1.13":
            ret = PkcsObjectIdentifier.SHA512WITH_RSA_ENCRYPTION
        elif value == "1.2.840.113549.1.1.14":
            ret = PkcsObjectIdentifier.SHA224WITH_RSA_ENCRYPTION
        elif value == "1.2.840.113549.1.3.1":
            ret = PkcsObjectIdentifier.DH_KEY_AGREE1MENT
        elif value == "1.2.840.113549.1.5.1":
            ret = PkcsObjectIdentifier.PBE_WITH_MD2AND_DES_CBC
        elif value == "1.2.840.113549.1.5.4":
            ret = PkcsObjectIdentifier.PBE_WITH_MD2AND_RC2CBC
        elif value == "1.2.840.113549.1.5.3":
            ret = PkcsObjectIdentifier.PBE_WITH_MD5AND_DES_CBC
        elif value == "1.2.840.113549.1.5.6":
            ret = PkcsObjectIdentifier.PBE_WITH_MD5AND_RC2CBC
        elif value == "1.2.840.113549.1.5.10":
            ret = PkcsObjectIdentifier.PBE_WITH_SHA1AND_DES_CBC
        elif value == "1.2.840.113549.1.5.11":
            ret = PkcsObjectIdentifier.PBE_WITH_SHA1AND_RC2CBC
        elif value == "1.2.840.113549.1.5.13":
            ret = PkcsObjectIdentifier.ID_PBE_S2
        elif value == "1.2.840.113549.1.5.12":
            ret = PkcsObjectIdentifier.ID_PBKDF2
        elif value == "1.2.840.113549.3.7":
            ret = PkcsObjectIdentifier.DES_EDE3CBC
        elif value == "1.2.840.113549.3.2":
            ret = PkcsObjectIdentifier.RC2CBC
        elif value == "1.2.840.113549.2.2":
            ret = PkcsObjectIdentifier.MD2
        elif value == "1.2.840.113549.2.4":
            ret = PkcsObjectIdentifier.MD4
        elif value == "1.2.840.113549.2.5":
            ret = PkcsObjectIdentifier.MD5
        elif value == "1.2.840.113549.2.7":
            ret = PkcsObjectIdentifier.ID_HMAC_WITH_SHA1
        elif value == "1.2.840.113549.2.8":
            ret = PkcsObjectIdentifier.ID_HMAC_WITH_SHA224
        elif value == "1.2.840.113549.2.9":
            ret = PkcsObjectIdentifier.ID_HMAC_WITH_SHA256
        elif value == "1.2.840.113549.2.10":
            ret = PkcsObjectIdentifier.ID_HMAC_WITH_SHA384
        elif value == "1.2.840.113549.2.11":
            ret = PkcsObjectIdentifier.ID_HMAC_WITH_SHA512
        elif value == "1.2.840.113549.1.7.1":
            ret = PkcsObjectIdentifier.DATA
        elif value == "1.2.840.113549.1.7.2":
            ret = PkcsObjectIdentifier.SIGNED_DATA
        elif value == "1.2.840.113549.1.7.3":
            ret = PkcsObjectIdentifier.ENVELOPED_DATA
        elif value == "1.2.840.113549.1.7.4":
            ret = PkcsObjectIdentifier.SIGNED_AND_ENVELOPED_DATA
        elif value == "1.2.840.113549.1.7.5":
            ret = PkcsObjectIdentifier.DIGESTED_DATA
        elif value == "1.2.840.113549.1.7.6":
            ret = PkcsObjectIdentifier.ENCRYPTED_DATA
        elif value == "1.2.840.113549.1.9.1":
            ret = PkcsObjectIdentifier.PKCS9AT_EMAIL_ADDRESS
        elif value == "1.2.840.113549.1.9.2":
            ret = PkcsObjectIdentifier.PKCS9AT_UNSTRUCTURED_NAME
        elif value == "1.2.840.113549.1.9.3":
            ret = PkcsObjectIdentifier.PKCS9AT_CONTENT_TYPE
        elif value == "1.2.840.113549.1.9.4":
            ret = PkcsObjectIdentifier.PKCS9AT_MESSAGE_DIGEST
        elif value == "1.2.840.113549.1.9.5":
            ret = PkcsObjectIdentifier.PKCS9AT_SIGNING_TIME
        elif value == "1.2.840.113549.1.9.6":
            ret = PkcsObjectIdentifier.PKCS9AT_COUNTER_SIGNATURE
        elif value == "1.2.840.113549.1.9.7":
            ret = PkcsObjectIdentifier.PKCS9AT_CHALLENGE_PASSWORD
        elif value == "1.2.840.113549.1.9.8":
            ret = PkcsObjectIdentifier.PKCS9AT_UNSTRUCTURED_ADDRESS
        elif value == "1.2.840.113549.1.9.9":
            ret = PkcsObjectIdentifier.PKCS9AT_EXTENDED_CERTIFICATE_ATTRIBUTES
        elif value == "1.2.840.113549.1.9.13":
            ret = PkcsObjectIdentifier.PKCS9AT_SIGNING_DESCRIPTION
        elif value == "1.2.840.113549.1.9.14":
            ret = PkcsObjectIdentifier.PKCS9AT_EXTENSION_REQUEST
        elif value == "1.2.840.113549.1.9.15":
            ret = PkcsObjectIdentifier.PKCS9AT_SMIME_CAPABILITIES
        elif value == "1.2.840.113549.1.9.16":
            ret = PkcsObjectIdentifier.ID_SMIME
        elif value == "1.2.840.113549.1.9.20":
            ret = PkcsObjectIdentifier.PKCS9AT_FRIENDLY_NAME
        elif value == "1.2.840.113549.1.9.21":
            ret = PkcsObjectIdentifier.PKCS9AT_LOCAL_KEY_ID
        elif value == "1.2.840.113549.1.9.22.1":
            ret = PkcsObjectIdentifier.X509CERTIFICATE
        elif value == "1.2.840.113549.1.9.22.2":
            ret = PkcsObjectIdentifier.SDSI_CERTIFICATE
        elif value == "1.2.840.113549.1.9.23.1":
            ret = PkcsObjectIdentifier.X509CRL
        elif value == "1.2.840.113549.1.9.16.3":
            ret = PkcsObjectIdentifier.ID_ALG
        elif value == "1.2.840.113549.1.9.16.3.5":
            ret = PkcsObjectIdentifier.ID_ALG_ESDH
        elif value == "1.2.840.113549.1.9.16.3.6":
            ret = PkcsObjectIdentifier.ID_ALG_CMS3DES_WRAP
        elif value == "1.2.840.113549.1.9.16.3.7":
            ret = PkcsObjectIdentifier.ID_ALG_CMS_RC2WRAP
        elif value == "1.2.840.113549.1.9.16.3.9":
            ret = PkcsObjectIdentifier.ID_ALG_PWRI_KEK
        elif value == "1.2.840.113549.1.9.16.3.10":
            ret = PkcsObjectIdentifier.ID_ALG_SSDH
        elif value == "1.2.840.113549.1.9.16.3.14":
            ret = PkcsObjectIdentifier.ID_RSA_KEM
        elif value == "1.2.840.113549.1.9.15.1":
            ret = PkcsObjectIdentifier.PREFER_SIGNED_DATA
        elif value == "1.2.840.113549.1.9.15.2":
            ret = PkcsObjectIdentifier.CANNOT_DECRYPT_ANY
        elif value == "1.2.840.113549.1.9.15.3":
            ret = PkcsObjectIdentifier.SMIME_CAPABILITIES_VERSIONS
        elif value == "1.2.840.113549.1.9.16.2.1":
            ret = PkcsObjectIdentifier.ID_AA_RECEIPT_REQUEST
        elif value == "1.2.840.113549.1.9.16.1.2":
            ret = PkcsObjectIdentifier.ID_CT_AUTH_DATA
        elif value == "1.2.840.113549.1.9.16.1.4":
            ret = PkcsObjectIdentifier.ID_CT_TST_INFO
        elif value == "1.2.840.113549.1.9.16.1.9":
            ret = PkcsObjectIdentifier.ID_CT_COMPRESSED_DATA
        elif value == "1.2.840.113549.1.9.16.1.23":
            ret = PkcsObjectIdentifier.ID_CT_AUTH_ENVELOPED_DATA
        elif value == "1.2.840.113549.1.9.16.1.31":
            ret = PkcsObjectIdentifier.ID_CT_TIMESTAMPED_DATA
        elif value == "1.2.840.113549.1.9.16.6.1":
            ret = PkcsObjectIdentifier.ID_CTI_ETS_PROOF_OF_ORIGIN
        elif value == "1.2.840.113549.1.9.16.6.2":
            ret = PkcsObjectIdentifier.ID_CTI_ETS_PROOF_OF_RECEIPT
        elif value == "1.2.840.113549.1.9.16.6.3":
            ret = PkcsObjectIdentifier.ID_CTI_ETS_PROOF_OF_DELIVERY
        elif value == "1.2.840.113549.1.9.16.6.4":
            ret = PkcsObjectIdentifier.ID_CTI_ETS_PROOF_OF_SENDER
        elif value == "1.2.840.113549.1.9.16.6.5":
            ret = PkcsObjectIdentifier.ID_CTI_ETS_PROOF_OF_APPROVAL
        elif value == "1.2.840.113549.1.9.16.6.6":
            ret = PkcsObjectIdentifier.ID_CTI_ETS_PROOF_OF_CREATION
        elif value == "1.2.840.113549.1.9.16.2.4":
            ret = PkcsObjectIdentifier.ID_AA_CONTENT_HINT
        elif value == "1.2.840.113549.1.9.16.2.5":
            ret = PkcsObjectIdentifier.ID_AA_MSG_SIG_DIGEST
        elif value == "1.2.840.113549.1.9.16.2.10":
            ret = PkcsObjectIdentifier.ID_AA_CONTENT_REFERENCE
        elif value == "1.2.840.113549.1.9.16.2.11":
            ret = PkcsObjectIdentifier.ID_AA_ENCRYP_KEY_PREF
        elif value == "1.2.840.113549.1.9.16.2.12":
            ret = PkcsObjectIdentifier.ID_AA_SIGNING_CERTIFICATE
        elif value == "1.2.840.113549.1.9.16.2.47":
            ret = PkcsObjectIdentifier.ID_AA_SIGNING_CERTIFICATE_V2
        elif value == "1.2.840.113549.1.9.16.2.7":
            ret = PkcsObjectIdentifier.ID_AA_CONTENT_IDENTIFIER
        elif value == "1.2.840.113549.1.9.16.2.14":
            ret = PkcsObjectIdentifier.ID_AA_SIGNATURE_TIME_STAMP_TOKEN
        elif value == "1.2.840.113549.1.9.16.2.15":
            ret = PkcsObjectIdentifier.ID_AA_ETS_SIG_POLICY_ID
        elif value == "1.2.840.113549.1.9.16.2.16":
            ret = PkcsObjectIdentifier.ID_AA_ETS_COMMITMENT_TYPE
        elif value == "1.2.840.113549.1.9.16.2.17":
            ret = PkcsObjectIdentifier.ID_AA_ETS_SIGNER_LOCATION
        elif value == "1.2.840.113549.1.9.16.2.18":
            ret = PkcsObjectIdentifier.ID_AA_ETS_SIGNER_ATTR
        elif value == "1.2.840.113549.1.9.16.2.19":
            ret = PkcsObjectIdentifier.ID_AA_ETS_OTHER_SIG_CERT
        elif value == "1.2.840.113549.1.9.16.2.20":
            ret = PkcsObjectIdentifier.ID_AA_ETS_CONTENT_TIMESTAMP
        elif value == "1.2.840.113549.1.9.16.2.21":
            ret = PkcsObjectIdentifier.ID_AA_ETS_CERTIFICATE_REFS
        elif value == "1.2.840.113549.1.9.16.2.22":
            ret = PkcsObjectIdentifier.ID_AA_ETS_REVOCATION_REFS
        elif value == "1.2.840.113549.1.9.16.2.23":
            ret = PkcsObjectIdentifier.ID_AA_ETS_CERT_VALUES
        elif value == "1.2.840.113549.1.9.16.2.24":
            ret = PkcsObjectIdentifier.ID_AA_ETS_REVOCATION_VALUES
        elif value == "1.2.840.113549.1.9.16.2.25":
            ret = PkcsObjectIdentifier.ID_AA_ETS_ESC_TIME_STAMP
        elif value == "1.2.840.113549.1.9.16.2.26":
            ret = PkcsObjectIdentifier.ID_AA_ETS_CERT_CRL_TIMESTAMP
        elif value == "1.2.840.113549.1.9.16.2.27":
            ret = PkcsObjectIdentifier.ID_AA_ETS_ARCHIVE_TIMESTAMP
        elif value == "1.2.840.113549.1.9.16.5.1":
            ret = PkcsObjectIdentifier.ID_SPQ_ETS_URI
        elif value == "1.2.840.113549.1.9.16.5.2":
            ret = PkcsObjectIdentifier.ID_SPQ_ETS_U_NOTICE
        elif value == "1.2.840.113549.1.12.10.1.1":
            ret = PkcsObjectIdentifier.KEY_BAG
        elif value == "1.2.840.113549.1.12.10.1.2":
            ret = PkcsObjectIdentifier.PKCS8SHROUDED_KEY_BAG
        elif value == "1.2.840.113549.1.12.10.1.3":
            ret = PkcsObjectIdentifier.CERT_BAG
        elif value == "1.2.840.113549.1.12.10.1.4":
            ret = PkcsObjectIdentifier.CRL_BAG
        elif value == "1.2.840.113549.1.12.10.1.5":
            ret = PkcsObjectIdentifier.SECRET_BAG
        elif value == "1.2.840.113549.1.12.10.1.6":
            ret = PkcsObjectIdentifier.SAFE_CONTENTS_BAG
        elif value == "1.2.840.113549.1.12.1.1":
            ret = PkcsObjectIdentifier.PBE_WITH_SHA_AND128BIT_RC4
        elif value == "1.2.840.113549.1.12.1.2":
            ret = PkcsObjectIdentifier.PBE_WITH_SHA_AND40BIT_RC4
        elif value == "1.2.840.113549.1.12.1.3":
            ret = PkcsObjectIdentifier.PBE_WITH_SHA_AND3KEY_TRIPLE_DES_CBC
        elif value == "1.2.840.113549.1.12.1.4":
            ret = PkcsObjectIdentifier.PBE_WITH_SHA_AND2KEY_TRIPLE_DES_CBC
        elif value == "1.2.840.113549.1.12.1.5":
            ret = PkcsObjectIdentifier.PBE_WITH_SHA_AND128BIT_RC2CBC
        elif value == "1.2.840.113549.1.12.1.6":
            ret = PkcsObjectIdentifier.PBEWITH_SHA_AND40BIT_RC2CBC
        else:
            ret = PkcsObjectIdentifier.NONE
        return ret



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import random
import hashlib
from .enums import Authentication, Security
from .GXDLMSChipperingStream import GXDLMSChipperingStream
from .GXByteBuffer import GXByteBuffer
from .CountType import CountType
from .GXDLMSChippering import GXDLMSChippering
from .objects.enums.CertificateType import CertificateType
from .ecdsa.GXEcdsa import GXEcdsa
from .AesGcmParameter import AesGcmParameter


class GXSecure:
    #
    #      * Chipher text.
    #      *
    #      * @param settings
    #      * DLMS settings.
    #      * @param cipher
    #      * Chipher settings.
    #      * @param ic
    #      * IC
    #      * @param data
    #      * Text to chipher.
    #      * @param secret
    #      * Secret.
    #      * @return Chiphered text.
    # pylint: disable=too-many-arguments, too-many-locals
    @classmethod
    def secure(cls, settings, cipher, ic, data, secret):
        if isinstance(secret, str):
            secret = secret.encode()
        d = []
        s = []
        if settings.authentication == Authentication.HIGH:
            len_ = len(data)
            if (len_ % 16) != 0:
                len_ += 16 - (len(data) % 16)

            if len(secret) > len(data):
                len_ = len(secret)
            if len_ % 16 != 0:
                len_ += 16 - (len(secret) % 16)
            s = bytearray(len_)
            d = bytearray(len_)
            s[0 : len(secret)] = secret[0:]
            d[0 : len(data)] = data[0:]
            pos = 0
            while pos < len(d) / 16:
                GXDLMSChipperingStream.aes1Encrypt(d, pos * 16, s)
                pos += 1
            return d
        #  Get server Challenge.
        challenge = GXByteBuffer()
        #  Get shared secret
        if settings.authentication == Authentication.HIGH_GMAC:
            challenge.set(data)
        elif settings.authentication in (
            Authentication.HIGH_SHA256,
            Authentication.HIGH_ECDSA,
        ):
            challenge.set(secret)
        else:
            challenge.set(data)
            challenge.set(secret)
        d = challenge.array()
        if settings.authentication == Authentication.HIGH_MD5:
            md = hashlib.md5()
            md.update(d)
            d = md.digest()
        elif settings.authentication == Authentication.HIGH_SHA1:
            md = hashlib.sha1()
            md.update(d)
            d = md.digest()
        elif settings.authentication == Authentication.HIGH_SHA256:
            md = hashlib.sha256()
            md.update(d)
            d = md.digest()
        elif settings.authentication == Authentication.HIGH_GMAC:
            #  SC is always Security.Authentication.
            p = AesGcmParameter(
                0, secret, cipher.blockCipherKey, cipher.authenticationKey
            )
            p.security = Security.AUTHENTICATION
            p.securitySuite = cipher.securitySuite
            p.invocationCounter = ic
            p.type_ = CountType.TAG
            challenge.clear()
            challenge.setUInt8(Security.AUTHENTICATION | cipher.securitySuite)
            challenge.setUInt32(p.invocationCounter)
            challenge.set(GXDLMSChippering.encryptAesGcm(p, d))
            d = challenge.array()
        elif settings.authentication == Authentication.HIGH_ECDSA:
            key = settings.cipher.signingKeyPair
            if key:
                key = key[1]

            pub = settings.cipher.signingKeyPair
            if pub:
                pub = pub[0]
            if not key:
                key = settings.getKey(
                    CertificateType.DIGITAL_SIGNATURE, settings.cipher.systemTitle, True
                )
                settings.cipher.signingKeyPair = (pub, key)
            if not pub:
                pub = settings.getKey(
                    CertificateType.DIGITAL_SIGNATURE, settings.sourceSystemTitle, False
                )
                settings.cipher.signingKeyPair = (pub, key)
            if not key:
                raise ValueError("Signing key is not set.")
            print("Private signed key: " + key.toHex())
            print("Public signed key: " + key.getPublicKey().toHex())
            sig = GXEcdsa(key)
            d = sig.sign(secret)
        return d

    #
    #      * Generates challenge.
    #      *
    #      * @param authentication
    #      * Used authentication.
    #      * @return Generated challenge.
    #
    @classmethod
    def generateChallenge(cls):
        #  Random challenge is 8 to 64 bytes.
        #  Texas Instruments accepts only 16 byte long challenge.
        #  For this reason challenge size is 16 bytes at the moment.
        len_ = 16
        result = [None] * len_
        pos = 0
        while pos != len_:
            result[pos] = random.randint(0, 255)
            pos += 1
        return result



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
#
#  --------------------------------------------------------------------------
#


class GXBitString:
    # pylint: disable=import-outside-toplevel
    def __init__(self, value=None, pad_count=None):
        from .internal._GXCommon import _GXCommon

        self.__padBits = 0
        self.value = bytearray()

        if isinstance(value, str):
            self.pad_bits = 8 - (len(value) % 8)
            if self.pad_bits == 8:
                self.pad_bits = 0
            bit_str = value + "0" * self.pad_bits
            self.value = bytearray(
                int(bit_str[i : i + 8], 2) for i in range(0, len(bit_str), 8)
            )

        elif isinstance(value, (bytes, bytearray)) and pad_count is not None:
            if pad_count < 0 or pad_count > 7:
                raise ValueError("PadCount must be in the range 0 to 7")
            self.value = bytearray(value)
            self.pad_bits = pad_count

        elif isinstance(value, (bytes, bytearray)):
            if len(value) == 0:
                raise ValueError("value cannot be empty")
            self.pad_bits = value[0]
            if self.pad_bits < 0 or self.pad_bits > 7:
                raise ValueError("PadCount must be in the range 0 to 7")
            self.value = bytearray(value[1:])

        elif isinstance(value, int) and pad_count is not None:
            self.pad_bits = pad_count % 8
            length = (pad_count // 8) + (1 if self.pad_bits else 0)
            self.value = bytearray()
            for _ in range(length):
                self.value.append(_GXCommon.swapBits(value & 0xFF))
                value >>= 8

    @property
    def padBits(self):
        """Number of extra bits at the end of the string."""
        return self.__padBits

    @property
    def length(self):
        return 0 if not self.value else (8 * len(self.value)) - self.pad_bits

    def __str__(self):
        return self.toString(False)

    # pylint: disable=import-outside-toplevel
    def toString(self, show_bits=False):
        from .internal._GXCommon import _GXCommon

        if not self.value:
            return ""
        bits = "".join(_GXCommon.toBitString(byte, 8) for byte in self.value)
        bits = bits[: len(bits) - self.pad_bits]

        return f"{len(bits)} bit {bits}" if show_bits else bits

    def toInteger(self):
        from .internal._GXCommon import _GXCommon

        result = 0
        pos = 0
        for byte in self.value:
            result |= _GXCommon.swapBits(byte) << pos
            pos += 8
        return result



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http:#www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
class GXDLMSContextInformationTable:
    '''
    Contains the context information associated to each CID extension field.
    '''

    __cID = ""
    "Corresponds to the 4-bit context information used for source and destination addresses (SCI, DCI)."

    __context = bytearray()
    "Context."

    __compression = None
    "Indicates if the context is valid for use in compression."

    __validLifetime = None
    "Remaining time in minutes during which the context information table is considered valid. It is updated upon reception of the advertised context."

    @property
    def cID(self):
        '''
        Corresponds to the 4-bit context information used for source and destination addresses (SCI, DCI).
        '''
        return self.__cID

    @cID.setter
    def cID(self, value):
        self.__cID = value
    @property
    def context(self):
        '''
        Context.
        '''
        return self.__context

    @context.setter
    def context(self, value):
        self.__context = value
    @property
    def compression(self):
        '''
        Indicates if the context is valid for use in compression.
        '''
        return self.__compression

    @compression.setter
    def compression(self, value):
        self.__compression = value
    @property
    def validLifetime(self):
        '''
        Remaining time in minutes during which the context information table is considered valid. It is updated upon reception of the advertised context.
        '''
        return self.__validLifetime

    @validLifetime.setter
    def validLifetime(self, value):
        self.__validLifetime = value



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from __future__ import print_function
from .enums import (
    Priority,
    ServiceClass,
    InterfaceType,
    Authentication,
    Standard,
    HdlcFrameType,
    Conformance,
)
from .enums.DateTimeSkips import DateTimeSkips
from .ConnectionState import ConnectionState
from .objects.GXDLMSObjectCollection import GXDLMSObjectCollection
from .GXHdlcSettings import GXHdlcSettings
from .IGXCryptoNotifier import IGXCryptoNotifier
from .objects.enums.CertificateType import CertificateType
from .GXCryptoKeyParameter import GXCryptoKeyParameter
from .enums.ObjectType import ObjectType


# This class includes DLMS communication settings.
# pylint: disable=bad-option-value,too-many-public-methods,too-many-instance-attributes,old-style-class
class GXDLMSSettings:
    #
    # Server sender frame sequence starting number.
    #
    __SERVER_START_SENDER_FRAME_SEQUENCE = 0x1E

    #
    # Server receiver frame sequence starting number.
    #
    __SERVER_START_RECEIVER_FRAME_SEQUENCE = 0xFE

    #
    # Client sender frame sequence starting number.
    #
    __CLIENT_START_SENDER_FRAME_SEQUENCE = 0xFE

    #
    # Client receiver frame sequence starting number.
    #
    __CLIENT_START_RCEIVER_FRAME_SEQUENCE = 0xE

    #
    # DLMS version number.
    #
    __DLMS_VERSION = 6
    __MAX_RECEIVE_PDU_SIZE = 0xFFFF

    @classmethod
    def getInitialConformance(cls, useLN):
        if useLN:
            return (
                Conformance.GENERAL_PROTECTION
                | Conformance.BLOCK_TRANSFER_WITH_ACTION
                | Conformance.BLOCK_TRANSFER_WITH_SET_OR_WRITE
                | Conformance.BLOCK_TRANSFER_WITH_GET_OR_READ
                | Conformance.SET
                | Conformance.SELECTIVE_ACCESS
                | Conformance.ACTION
                | Conformance.MULTIPLE_REFERENCES
                | Conformance.GET
                | Conformance.ACCESS
            )
        return (
            Conformance.INFORMATION_REPORT
            | Conformance.READ
            | Conformance.UN_CONFIRMED_WRITE
            | Conformance.WRITE
            | Conformance.PARAMETERIZED_ACCESS
            | Conformance.MULTIPLE_REFERENCES
        )

    #
    # Constructor.
    #
    def __init__(self, isServer, parent):
        self.customChallenges = False
        self.ctoSChallenge = None
        self.stoCChallenge = None
        self.sourceSystemTitle = None
        self.invokeId = 0x1
        self.longInvokeID = 0x1
        self.priority = Priority.HIGH
        self.serviceClass = ServiceClass.CONFIRMED
        self.clientAddress = 0
        self.serverAddress = 0
        self.pushClientAddress = 0
        self.serverAddressSize = 0
        self.__useLogicalNameReferencing = True
        self.interfaceType = InterfaceType.HDLC
        self.authentication = Authentication.NONE
        self.password = None
        self.kek = None
        self.count = 0
        self.index = 0
        self.targetEphemeralKey = None
        self.dlmsVersion = self.__DLMS_VERSION
        self.connected = ConnectionState.NONE
        self.allowAnonymousAccess = False
        self.maxPduSize = self.__MAX_RECEIVE_PDU_SIZE
        self.maxServerPDUSize = self.__MAX_RECEIVE_PDU_SIZE
        self.startingPacketIndex = 1
        # Gets current block index.
        self.blockIndex = 1
        self.cipher = None
        self.blockNumberAck = 0
        self.protocolVersion = None
        self.isServer = isServer
        self.objects = GXDLMSObjectCollection()
        self.hdlc = GXHdlcSettings()
        self.gateway = None
        self.proposedConformance = GXDLMSSettings.getInitialConformance(
            self.__useLogicalNameReferencing
        )
        self.receiverFrame = 0
        self.senderFrame = 0
        self.resetFrameSequence()
        self.gbtWindowSize = 1
        self.userId = -1
        # Quality of service.
        self.qualityOfService = 0
        self.useUtc2NormalTime = False
        self.increaseInvocationCounterForGMacAuthentication = False
        self.dateTimeSkips = DateTimeSkips.NONE
        self.standard = Standard.DLMS
        self.negotiatedConformance = Conformance.NONE
        self.invokeID = 0
        self.command = 0
        self.commandType = 0
        self.useCustomChallenge = False
        self.preEstablishedSystemTitle = None
        # Assigned association for the server.
        self.__assignedAssociation = None

        ###Event listeners.
        self.__listeners = []
        if isinstance(parent, IGXCryptoNotifier):
            self.__listeners.append(parent)

    #
    # Client to Server challenge.
    #
    def getCtoSChallenge(self):
        return self.ctoSChallenge

    #
    # Notify client from decrypted PDU.
    #
    def _onPduEventHandler(self, complete, data):
        for it in self.__listeners:
            it.onPduEventHandler(it, complete, data)

    #
    # @param value
    # Client to Server challenge.
    #
    def setCtoSChallenge(self, value):
        if not self.customChallenges or self.ctoSChallenge is None:
            self.ctoSChallenge = value

    #
    # Server to Client challenge.
    #
    def getStoCChallenge(self):
        return self.stoCChallenge

    #
    # @param value
    # Server to Client challenge.
    #
    def setStoCChallenge(self, value):
        if not self.customChallenges or self.stoCChallenge is None:
            self.stoCChallenge = value

    #
    # Is connection accepted.
    #
    def acceptConnection(self):
        return (
            self.connected != ConnectionState.NONE
            or self.allowAnonymousAccess
            or (self.cipher and self.cipher.sharedSecret)
        )

    #
    # Reset frame sequence.
    #
    def resetFrameSequence(self):
        if self.isServer:
            self.senderFrame = self.__SERVER_START_SENDER_FRAME_SEQUENCE
            self.receiverFrame = self.__SERVER_START_RECEIVER_FRAME_SEQUENCE
        else:
            self.senderFrame = self.__CLIENT_START_SENDER_FRAME_SEQUENCE
            self.receiverFrame = self.__CLIENT_START_RCEIVER_FRAME_SEQUENCE

    # pylint: disable=too-many-return-statements
    def checkFrame(self, frame_, xml):
        #  If notify
        if frame_ == 0x13:
            return True

        #  If U frame.
        if (frame_ & HdlcFrameType.U_FRAME) == HdlcFrameType.U_FRAME:
            if frame_ == 0x93:
                isEcho = (
                    not self.isServer
                    and frame_ == 0x93
                    and self.senderFrame in (0x10, 0xFE)
                    and self.receiverFrame == 0xE
                )
                self.resetFrameSequence()
                return not isEcho
            if frame_ == 0x73 and not self.isServer:
                return self.senderFrame == 0xFE and self.receiverFrame == 0xE
            return True
        #  If S -frame.
        if (frame_ & HdlcFrameType.S_FRAME) == HdlcFrameType.S_FRAME:
            # If echo.
            if frame_ == (self.senderFrame & 0xF1):
                return False
            self.receiverFrame = self.increaseReceiverSequence(self.receiverFrame)
            return True
        #  Handle I-frame.
        if (self.senderFrame & 0x1) == 0:
            expected = GXDLMSSettings.increaseReceiverSequence(
                GXDLMSSettings.increaseSendSequence(self.receiverFrame)
            )
            if frame_ == expected:
                self.receiverFrame = frame_
                return True
            # If the final bit is not set.
            if frame_ == (expected & ~0x10) and self.hdlc.windowSizeRX != 1:
                self.receiverFrame = frame_
                return True
            # If Final bit is not set for the previous message.
            if (self.receiverFrame & 0x10) == 0 and self.hdlc.windowSizeRX != 1:
                expected = 0x10 | self.increaseSendSequence(self.receiverFrame)
                if frame_ == expected:
                    self.receiverFrame = frame_
                    return True
                # If the final bit is not set.
                if frame_ == (expected & ~0x10):
                    ReceiverFrame = frame_
                    return True
        else:
            expected = self.increaseSendSequence(self.receiverFrame) & 0xFF
            #  If answer for RR.
            if frame_ == expected:
                self.receiverFrame = frame_
                return True
            if frame_ == (expected & ~0x10):
                self.receiverFrame = frame_
                return True
            if self.hdlc.windowSizeRX != 1:
                # If HDLC window size is bigger than one.
                expected = self.increaseReceiverSequence(
                    self.increaseSendSequence(ReceiverFrame)
                )
                if frame_ == expected:
                    ReceiverFrame = frame_
                    return True

        # If try to find data from bytestream and not real communicating.
        if xml and (
            (not self.isServer and self.receiverFrame == 0xE)
            or (self.isServer and self.receiverFrame == 0xEE)
        ):
            ReceiverFrame = frame_
            return True
        print("Invalid HDLC Frame: " + hex(frame_) + " Expected: " + hex(expected))
        return False

    #
    # Increase receiver sequence.
    #
    # @param value
    # Frame value.
    # Increased receiver frame sequence.
    #
    @classmethod
    def increaseReceiverSequence(cls, value):
        return ((value & 0xFF) + 0x20 | 0x10 | value & 0xE) & 0xFF

    #
    # Increase sender sequence.
    #
    # @param value
    # Frame value.
    # Increased sender frame sequence.
    #
    @classmethod
    def increaseSendSequence(cls, value):
        return ((value & 0xF0 | (value + 0x2) & 0xE) & 0xFF) & 0xFF

    #
    # Generates I-frame.
    # @param first
    # Is this first packet.
    # Generated I-frame
    #
    def getNextSend(self, first):
        if first:
            self.senderFrame = self.increaseReceiverSequence(
                self.increaseSendSequence(int(self.senderFrame))
            )
        else:
            self.senderFrame = self.increaseSendSequence(int(self.senderFrame))
        return self.senderFrame & 0xFF

    #
    # Generates Receiver Ready S-frame.
    #
    def getReceiverReady(self):
        self.senderFrame = self.increaseReceiverSequence((self.senderFrame | 1))
        return self.senderFrame & 0xF1

    #
    # Generates Keep Alive S-frame.
    #
    def getKeepAlive(self):
        self.senderFrame = self.senderFrame | 1
        return self.senderFrame & 0xF1

    #
    # Gets starting block index in HDLC framing.  Default is One based,
    #      but some
    # meters use Zero based value.  Usually this is not used.
    #
    # Current block index.
    #
    def getStartingPacketIndex(self):
        return self.startingPacketIndex

    #
    # Set starting block index in HDLC framing.  Default is One based,
    #      but some
    # meters use Zero based value.  Usually this is not used.
    #
    # @param value
    # Zero based starting index.
    #
    def setStartingPacketIndex(self, value):
        self.startingPacketIndex = value
        self.resetBlockIndex()

    #
    # Sets current block index.
    #
    # @param value
    # Block index.
    #
    def setBlockIndex(self, value):
        self.blockIndex = value

    #
    # Block number acknowledged in GBT.
    #
    def getBlockNumberAck(self):
        return self.blockNumberAck

    #
    # @param value
    # Block number acknowledged in GBT.
    #
    def setBlockNumberAck(self, value):
        self.blockNumberAck = value

    #
    # Resets block index to default value.
    #
    def resetBlockIndex(self):
        self.blockIndex = self.startingPacketIndex
        self.blockNumberAck = 0

    #
    # Increases block index.
    #
    def increaseBlockIndex(self):
        self.blockIndex += 1

    # Is Logical Name Referencing used.
    # Don't use property. For some reason Python 2.7 doesn't call it in Rasbperry PI.
    def getUseLogicalNameReferencing(self):
        return self.__useLogicalNameReferencing

    # Is Logical Name Referencing used.
    # Don't use property. For some reason Python 2.7 doesn't call it in Rasbperry PI.
    def setUseLogicalNameReferencing(self, value):
        if self.__useLogicalNameReferencing != value:
            self.__useLogicalNameReferencing = value
            self.proposedConformance = GXDLMSSettings.getInitialConformance(
                self.__useLogicalNameReferencing
            )

    #
    # Invoke ID.
    #
    def getInvokeID(self):
        return self.invokeID

    #
    # @param value
    # update invoke ID.
    #
    def updateInvokeId(self, value):
        # pylint: disable=bad-option-value,redefined-variable-type
        if (value & 0x80) != 0:
            self.priority = Priority.HIGH
        else:
            self.priority = Priority.NORMAL
        if (value & 0x40) != 0:
            self.serviceClass = ServiceClass.CONFIRMED
        else:
            self.serviceClass = ServiceClass.UN_CONFIRMED
        self.invokeID = int((value & 0xF))

    #
    # @param value
    # Invoke ID.
    #
    def setInvokeID(self, value):
        if value > 0xF:
            raise ValueError("Invalid InvokeID")
        self.invokeID = int(value)

    #
    # Invoke ID.
    #
    def getLongInvokeID(self):
        return self.longInvokeID

    #
    # @param value
    # Invoke ID.
    #
    def setLongInvokeID(self, value):
        if value > 0xFFFFFFFF:
            raise ValueError("Invalid InvokeID")
        self.longInvokeID = value

    #
    # Source system title.
    #
    def getSourceSystemTitle(self):
        return self.sourceSystemTitle

    #
    # @param value
    # Source system title.
    #
    def setSourceSystemTitle(self, value):
        if not value or len(value) != 8:
            raise ValueError("Invalid client system title.")
        self.sourceSystemTitle = value

    #
    # Long data count.
    #
    def getCount(self):
        return self.count

    #
    # @param value
    # Data count.
    #
    def setCount(self, value):
        if value < 0:
            raise ValueError("Invalid count.")
        self.count = value

    #
    # Long data index.
    #
    def getIndex(self):
        return self.index

    #
    # @param value
    # Long data index
    #
    def setIndex(self, value):
        if value < 0:
            raise ValueError("Invalid Index.")
        self.index = value

    # Encrypt or decrypt the data using external Hardware Security Module.
    def crypt(self, certificateType, Data, encrypt, keyType):
        for it in self.__listeners:
            if isinstance(it, IGXCryptoNotifier):
                args = GXCryptoKeyParameter()
                args.keyType = keyType
                args.encrypt = encrypt
                args.systemTitle = self.cipher.systemTitle
                args.recipientSystemTitle = self.sourceSystemTitle
                args.certificateType = certificateType
                args.invocationCounter = self.cipher.invocationCounter
                args.securitySuite = self.cipher.securitySuite
                args.securityPolicy = self.cipher.securityPolicy
                if encrypt:
                    args.plainText = Data
                else:
                    args.encrypted = Data
                args.authenticationKey = self.cipher.authenticationKey
                if (
                    len(self.cipher.dedicatedKey) == 16
                    and (self.connected & ConnectionState.DLMS) != 0
                ):
                    args.blockCipherKey = self.cipher.dedicatedKey
                else:
                    args.blockCipherKey = self.cipher.blockCipherKey
                it.crypto(it, args)
                if encrypt:
                    return args.encrypted
                return args.plainText
            return None

    # pylint: disable=import-outside-toplevel, too-many-nested-blocks
    def getKey(self, certificateType, systemTitle, encrypt):
        from .asn.enums.KeyUsage import KeyUsage
        from .objects.GXDLMSAssociationLogicalName import GXDLMSAssociationLogicalName

        for it in self.__listeners:
            if isinstance(it, IGXCryptoNotifier):
                if certificateType == CertificateType.DIGITAL_SIGNATURE:
                    if encrypt:
                        if self.cipher.signingKeyPair and self.cipher.signingKeyPair[1]:
                            return self.cipher.signingKeyPair[1]
                    elif self.cipher.signingKeyPair and self.cipher.signingKeyPair[0]:
                        return self.cipher.signingKeyPair[0]
                elif certificateType == CertificateType.KEY_AGREEMENT:
                    if encrypt:
                        if (
                            self.cipher.keyAgreementKeyPair
                            and self.cipher.keyAgreementKeyPair[1]
                        ):
                            return self.cipher.KeyAgreementKeyPair[1]
                    elif (
                        self.cipher.keyAgreementKeyPair
                        and self.cipher.keyAgreementKeyPair[0]
                    ):
                        return self.cipher.keyAgreementKeyPair[0]
                # Find keys from security setup object.
                if (
                    isinstance(self.__assignedAssociation, GXDLMSAssociationLogicalName)
                    and systemTitle
                ):
                    ss = self.__assignedAssociation.objectList.findByLN(
                        ObjectType.SECURITY_SETUP,
                        self.__assignedAssociation.securitySetupReference,
                    )
                    if ss:
                        if certificateType == CertificateType.DIGITAL_SIGNATURE:
                            cert = ss.ServerCertificates.FindBySystemTitle(
                                systemTitle, KeyUsage.DIGITAL_SIGNATURE
                            )
                            if cert and ss.signingKey:
                                if encrypt:
                                    if ss.signingKey[1]:
                                        return ss.signingKey[1]
                                elif cert.publicKey:
                                    return cert.publicKey
                        elif certificateType == CertificateType.KEY_AGREEMENT:
                            cert = ss.ServerCertificates.FindBySystemTitle(
                                systemTitle, KeyUsage.KEY_AGREEMENT
                            )
                            if cert and ss.keyAgreementKey:
                                if encrypt:
                                    if ss.keyAgreementKey[1]:
                                        return ss.KeyAgreementKey[1]
                                elif cert.publicKey:
                                    return cert.publicKey

                args = GXCryptoKeyParameter()
                args.encrypt = encrypt
                args.securitySuite = self.cipher.securitySuite
                args.certificateType = certificateType
                args.systemTitle = systemTitle
                if it.onKey:
                    it.onKey(it, args)
                if encrypt:
                    return args.privateKey
                return args.publicKey
        return None



#
# --------------------------------------------------------------------------
#  Gurux Ltd
#
#
#
# Filename:        $HeadURL$
#
# Version:         $Revision$,
#                  $Date$
#                  $Author$
#
# Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#  DESCRIPTION
#
# This file is a part of Gurux Device Framework.
#
# Gurux Device Framework is Open Source software you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation version 2 of the License.
# Gurux Device Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# More information of Gurux products: https:#www.gurux.org
#
# This code is licensed under the GNU General Public License v2.
# Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
import os
from ..asn.enums.CertificateVersion import CertificateVersion
from ..asn.enums.X9ObjectIdentifier import X9ObjectIdentifier
from ..ecdsa.enums.Ecc import Ecc
from ..objects.enums.CertificateType import CertificateType
from ..internal._GXCommon import _GXCommon
from ..GXByteBuffer import GXByteBuffer
from .GXAsn1Sequence import GXAsn1Sequence
from .GXAsn1ObjectIdentifier import GXAsn1ObjectIdentifier
from .X9ObjectIdentifierConverter import X9ObjectIdentifierConverter
from .GXAsn1Context import GXAsn1Context
from ..GXBitString import GXBitString
from .GXAsn1Converter import GXAsn1Converter
from ..GXInt8 import GXInt8
from ..ecdsa.GXDLMSCertificateException import GXDLMSCertificateException
from ..ecdsa.GXEcdsa import GXEcdsa
from .enums.PkcsType import PkcsType


class GXPkcs8:
    """
    Pkcs8 certification request. Private key is saved using this format.
    """

    __rawData = bytearray()
    """
    Loaded PKCS #8 certificate as a raw data.
    """

    __description = ""
    """
    Description is extra metadata that is saved to PEM file.
    """

    __version = CertificateVersion.VERSION1
    """
    Private key version.
    """

    __algorithm = X9ObjectIdentifier.ID_EC_PUBLIC_KEY
    """
    Algorithm.
    """

    __privateKey = None
    """
    Private key.
    """

    __publicKey = None
    """
    Public key.
    """

    @property
    def algorithm(self):
        """
        Algorithm.
        """
        return self.__algorithm

    @property
    def description(self):
        """
        Description is extra metadata that is saved to PEM file.
        """
        return self.__description

    @description.setter
    def description(self, value):
        self.__description = value

    @property
    def version(self):
        """
        Private key version.
        """
        return self.__version

    @version.setter
    def version(self, value):
        self.__version = value

    @property
    def privateKey(self):
        """
        Private key.
        """
        return self.__privateKey

    @property
    def publicKey(self):
        """
        Public key.
        """
        return self.__publicKey

    @property
    def encoded(self):
        if self.__rawData:
            return self.__rawData
        d = GXAsn1Sequence()
        d.append(GXInt8(self.version))
        d1 = GXAsn1Sequence()
        d1.append(
            GXAsn1ObjectIdentifier(
                X9ObjectIdentifierConverter.getString(self.__algorithm)
            )
        )
        if self.publicKey.scheme == Ecc.P256:
            alg = GXAsn1ObjectIdentifier("1.2.840.10045.3.1.7")
        else:
            alg = GXAsn1ObjectIdentifier("1.3.132.0.34")
        d1.append(alg)
        d.append(d1)
        d2 = GXAsn1Sequence()
        d2.append(GXInt8(1))
        d2.append(self.privateKey.rawValue)
        d3 = GXAsn1Context()
        d3.index = 1
        d3.append(GXBitString(self.publicKey.rawValue, 0))
        d2.append(d3)
        d.append(GXAsn1Converter.toByteArray(d2))
        self.__rawData = GXAsn1Converter.toByteArray(d)
        return self.__rawData

    def __init__(self, pair=None):
        """
        Constructor.

            Parameters:
                pair: Private key or private/public key pair.
        """
        if pair:
            self.__privateKey = pair[1]
            self.__publicKey = pair[0]

    @classmethod
    def getFilePath(cls, scheme, certificateType, systemTitle):
        """
        Returns default file path.

            Parameters:
                scheme: Used scheme.
                certificateType: Certificate type.
                systemTitle: System title.

            Returns:
                File path.
        """
        if certificateType == CertificateType.DIGITAL_SIGNATURE:
            path = "D"
        elif certificateType == CertificateType.KEY_AGREEMENT:
            path = "A"
        elif certificateType == CertificateType.TLS:
            path = "T"
        else:
            raise ValueError("Unknown certificate type.")

        path += GXByteBuffer.hex(systemTitle, False) + ".pem"
        if scheme == Ecc.P256:
            path = os.path.join("Keys", path)
        else:
            path = os.path.join("Keys384", path)
        return path

    # pylint: disable=unused-private-member
    @classmethod
    def fromPem(cls, data):
        """
        Create PKCS #8 from PEM string.

            Parameters:
                data: PEM string.
        """
        START = "PRIVATE KEY-----\n"
        END = "-----END"
        data = data.replace("\r\n", "\n")
        start = data.find(START)
        if start == -1:
            raise ValueError("Invalid PEM file.")
        desc = None
        if start != len(START):
            desc = data[:start]
            DESCRIPTION = "#Description"
            # Check if there is a description metadata.
            descStart = desc.rfind(DESCRIPTION)
            if descStart != -1:
                descEnd = desc.find("\n", descStart, start)
                desc = desc[descStart + len(DESCRIPTION) : descEnd - len(DESCRIPTION)]
                desc = desc.strip()
            else:
                desc = None
        data = data[start + len(START) :]
        end = data.find(END)
        if end == -1:
            raise ValueError("Invalid PEM file.")

        cert = cls.fromDer(data[0:end])
        cert.__description = desc
        return cert

    @classmethod
    def fromHexString(cls, data):
        """
        Create PKCS 8 from hex string.

            Parameters:
                data: Hex string.

        Returns:
            PKCS 8
        """
        cert = GXPkcs8()
        cert.__init(GXByteBuffer.hexToBytes(data))
        return cert

    @classmethod
    def fromDer(cls, der):
        """
           Create PKCS #8 from DER Base64 encoded string.

               Parameters:
                   der: Base64 DER string.

        *
        * @return
        """
        der = der.replace("\r\n", "")
        der = der.replace("\n", "")
        cert = GXPkcs8()
        cert.__init(_GXCommon.fromBase64(der))
        return cert

    # pylint: disable=import-outside-toplevel, unused-private-member, protected-access
    def __init(self, data):
        """
        Initialize the private key from ASN.1 encoded data.
        :param data: Byte array containing ASN.1 encoded PKCS#8 data.
        """
        from gurux_dlms.ecdsa.GXPrivateKey import GXPrivateKey
        from gurux_dlms.ecdsa.GXPublicKey import GXPublicKey

        self.__rawData = data
        seq = GXAsn1Converter.fromByteArray(data)
        if len(seq) < 3:
            raise ValueError("Wrong number of elements in sequence.")

        if not isinstance(seq[0], int):
            cert_type = GXAsn1Converter.__getCertificateType(data, seq)
            if cert_type == PkcsType.PKCS10:
                raise GXDLMSCertificateException(
                    "Invalid Certificate. This is PKCS 10 certification request, not PKCS 8."
                )
            if cert_type == PkcsType.X509CERTIFICATE:
                raise GXDLMSCertificateException(
                    "Invalid Certificate. This is a PKCS x509 certificate, not PKCS 8."
                )
            raise GXDLMSCertificateException("Invalid Certificate Version.")

        self.version = CertificateVersion(seq[0])

        if isinstance(seq[1], (bytes, bytearray)):
            raise GXDLMSCertificateException(
                "Invalid Certificate. This looks more like a private key, not PKCS 8."
            )

        tmp = seq[1]
        self.__algorithm = X9ObjectIdentifierConverter.fromString(str(tmp[0]))
        self.__privateKey = GXPrivateKey.fromRawBytes(seq[2][1])
        if self.__privateKey is None:
            raise Exception("Invalid private key.")

        # Public key optional
        tmp2 = seq[2]
        if len(tmp2) > 2:
            self.__publicKey = GXPublicKey.fromRawBytes(tmp2[2][0].value)
            GXEcdsa.validate(self.__publicKey)
        else:
            self.__publicKey = self.__privateKey.getPublicKey()

    def __str__(self):
        bb = "PKCS #8:"
        bb += "Version: " + str(self.version) + "\n"
        bb += "Algorithm: " + str(self.algorithm) + "\n"
        bb += "PrivateKey: " + self.privateKey.toHex() + "\n"
        bb += "PublicKey: " + str(self.publicKey)
        return bb

    @classmethod
    def load(cls, path):
        """
        Load private key from the PEM file.

            Parameters:
                path: File path.

            Returns:
                Created GXPkcs8 object.
        """

        with open(path, "r", encoding="utf-8") as file:
            return cls.fromPem(file.read())

    def save(self, path):
        """
        Save private key to PEM file.

            Parameters:
                path: File path.
        """

        with open(path, "w", encoding="utf-8") as file:
            file.write(self.toPem())

    def toPem(self):
        """
           Private key in PEM format.
        *
        * @return Private key as in PEM string.
        """
        sb = ""
        if not self.privateKey:
            raise ValueError("Public or private key is not set.")
        if self.description:
            sb = "#Description"
            sb += self.description + "\r"
        sb += "-----BEGIN PRIVATE KEY-----\r"
        sb += self.toDer()
        sb += "\r-----END PRIVATE KEY-----"
        return sb

    def toDer(self):
        """
        Private key in DER format.

            Returns:
                Private key as in DER string.
        """
        return _GXCommon.toBase64(self.encoded)

    def __eq__(self, other):
        return isinstance(other, GXPkcs8) and self.privateKey == other.privateKey



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright =c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from ..GXIntFlag import GXIntFlag

class ClockStatus(GXIntFlag):
    """Defines Clock status."""
    #pylint: disable=too-few-public-methods

    # OK.
    OK = 0

    # Invalid value.
    INVALID_VALUE = 0x1

    # Doubtful b value.
    DOUBTFUL_VALUE = 0x2

    # Different clock base c.
    DIFFERENT_CLOCK_BASE = 0X4

    # Invalid clock status d.
    INVALID_CLOCK_STATUS = 0x8

    # Reserved.
    RESERVED2 = 0x10

    # Reserved.
    RESERVED3 = 0x20

    # Reserved.
    RESERVED4 = 0x40

    # Daylight saving active.
    DAYLIGHT_SAVE_ACTIVE = 0x80

    # Clock status is skipped.
    #pylint: disable=W0213
    SKIPPED = 0xFF



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http:#www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from gurux_dlms.asn.enums.ExtendedKeyUsage import ExtendedKeyUsage
from gurux_dlms.objects.enums.CertificateType import CertificateType


class GXCertificateRequest:
    """
    Certificate request
    """

    __certificateType = CertificateType.DIGITAL_SIGNATURE
    """
    Certificate type.
    """

    __extendedKeyUsage = ExtendedKeyUsage.NONE
    """
    Indicates the purpose for which the certified public key is used.
    """

    __certificate = None
    """
    Certificate Signing Request.
    """

    @property
    def certificateType(self):
        """
        Certificate type.
        """
        return self.__certificateType

    @certificateType.setter
    def certificateType(self, value):
        self.__certificateType = value

    @property
    def extendedKeyUsage(self):
        """
        Indicates the purpose for which the certified public key is used.
        """
        return self.__extendedKeyUsage

    @extendedKeyUsage.setter
    def extendedKeyUsage(self, value):
        self.__extendedKeyUsage = value

    @property
    def certificate(self):
        """
        Certificate Signing Request.
        """
        return self.__certificate

    @certificate.setter
    def certificate(self, value):
        self.__certificate = value

    def __init__(
        self, certificateType=CertificateType.DIGITAL_SIGNATURE, certificate=None
    ):
        """
        Constructor.
        """
        self.__certificate = certificate
        self.__certificateType = certificateType



#
# --------------------------------------------------------------------------
#  Gurux Ltd
#
#
#
# Filename:        $HeadURL$
#
# Version:         $Revision$,
#                  $Date$
#                  $Author$
#
# Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#  DESCRIPTION
#
# This file is a part of Gurux Device Framework.
#
# Gurux Device Framework is Open Source software you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation version 2 of the License.
# Gurux Device Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# More information of Gurux products: https:#www.gurux.org
#
# This code is licensed under the GNU General Public License v2.
# Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from gurux_dlms.ecdsa.enums.Ecc import Ecc
from gurux_dlms.GXByteBuffer import GXByteBuffer
from gurux_dlms.internal._GXCommon import _GXCommon

from gurux_dlms.asn.GXAsn1Converter import GXAsn1Converter
from gurux_dlms.asn.X9ObjectIdentifierConverter import X9ObjectIdentifierConverter
from gurux_dlms.asn.enums.X9ObjectIdentifier import X9ObjectIdentifier
from gurux_dlms.asn.GXAsn1Sequence import GXAsn1Sequence
from gurux_dlms.asn.GXAsn1ObjectIdentifier import GXAsn1ObjectIdentifier

# from gurux_dlms.GXDLMSTranslator import GXDLMSTranslator
from gurux_dlms.GXBitString import GXBitString


class GXPublicKey:
    """
    Public key.
    """

    _scheme = Ecc.P256
    """
    Used scheme.
    """

    __rawValue = bytearray()
    """
    Private key raw value.
    """

    __systemTitle = bytearray()
    """
    SystemTitle is an extra information that can be used in debugging.
    """

    @property
    def scheme(self):
        """
        Used scheme.
        """
        return self._scheme

    @property
    def rawValue(self):
        """
        Private key raw value.
        """
        return self.__rawValue

    @property
    def systemTitle(self):
        """
        SystemTitle is an extra information that can be used in debugging.
        """
        return self.__systemTitle

    @systemTitle.setter
    def systemTitle(self, value):
        self.__systemTitle = value

    @property
    def x(self):
        """
        X Coordinate.
        """
        pk = GXByteBuffer(self.__rawValue)
        size = int(pk.size / 2)
        return pk.subArray(1, size)

    @property
    def y(self):
        """
        Y Coordinate.
        """
        pk = GXByteBuffer(self.__rawValue)
        size = int(pk.size / 2)
        return pk.subArray(1 + size, size)

    # pylint: disable=unused-private-member
    @classmethod
    def fromRawBytes(cls, key):
        """
        Create the public key from raw bytes.

            Parameters:
                key: Raw data

            Returns:
                Public key.
        """
        value = GXPublicKey()
        if len(key) == 65:
            value._scheme = Ecc.P256
            value.__rawValue = key
        elif len(key) == 97:
            value._scheme = Ecc.P384
            value.__rawValue = key
        elif len(key) == 64:
            # Compression tag is not send in DLMS messages.
            value._scheme = Ecc.P256
            value.__rawValue = [65]
            value.__rawValue[0] = 4
            value.__rawValue[1:64] = key[0:64]
        elif len(key) == 96:
            # Compression tag is not send in DLMS messages.
            value._scheme = Ecc.P384
            value.__rawValue = [96]
            value.__rawValue[0] = 4
            value.__rawValue[1:95] = key[0:95]
        else:
            raise ValueError("Invalid public key.")
        return value

    @classmethod
    def fromDer(cls, der):
        """
        Create the public key from DER.

            Parameters:
                der: DER Base64 coded string.

            Returns:
                Public key.
        """
        der = der.replace("\r\n", "")
        der = der.replace("\n", "")
        value = GXPublicKey()
        key = _GXCommon.fromBase64(der)
        seq = GXAsn1Converter.fromByteArray(key)
        id_ = X9ObjectIdentifierConverter.fromString(str(seq[0][1]))
        if id_ == X9ObjectIdentifier.PRIME256V1:
            value._scheme = Ecc.P256
        elif id_ == X9ObjectIdentifier.SECP384R1:
            value._scheme = Ecc.P384
        else:
            if id_ == X9ObjectIdentifier.NONE:
                raise ValueError("Invalid public key " + str(seq[0]) + ".")
            raise ValueError("Invalid public key " + id_ + " " + str(seq[0]) + ".")
        if isinstance(seq[1], bytes):
            value.__rawValue = seq[1]
        else:
            # Open SSL PEM.
            value.__rawValue = seq[1].value
        return value

    @classmethod
    def fromPem(cls, pem):
        """
        Create the public key from PEM.

            Parameters:
                pem: PEM Base64 coded string.

            Returns:
                Public key.
        """
        pem = pem.replace("\r\n", "\n")
        START = "-----BEGIN PUBLIC KEY-----\n"
        END = "\n-----END PUBLIC KEY-----"
        index = pem.find(START)
        if index == -1:
            raise ValueError("Invalid PEM file.")
        pem = pem[index + len(START) :]
        index = pem.find(END)
        if index == -1:
            raise ValueError("Invalid PEM file.")
        return cls.fromDer(pem[0:index])

    @classmethod
    def load(cls, path):
        """
        Create the public key from PEM file.

            Parameters:
                path: Path to the PEM file.

            Returns
            Public key.
        """
        with open(path, "r", encoding="utf-8") as file:
            return cls.fromPem(file.read())

    def save(self, path):
        """
        Save public key to PEM file.

            Parameters:
                path: File path.
        """
        with open(path, "w", encoding="utf-8") as file:
            file.write(self.toPem())

    def toHex(self):
        """
        Returns the public key as a hex string.
        """
        return GXByteBuffer.hex(self.__rawValue)

    def toDer(self):
        """
        Get public key as DER format.
        """
        return _GXCommon.toBase64(self.toEncoded())

    def toEncoded(self):
        """
        Get public key as encoded format.
        """
        # Subject Public Key Info.
        d = GXAsn1Sequence()
        d1 = GXAsn1Sequence()
        d1.append(GXAsn1ObjectIdentifier("1.2.840.10045.2.1"))
        if self.scheme == Ecc.P256:
            d1.append(GXAsn1ObjectIdentifier("1.2.840.10045.3.1.7"))
        elif self.scheme == Ecc.P384:
            d1.append(GXAsn1ObjectIdentifier("1.3.132.0.34"))
        else:
            raise ValueError("Invalid ECC scheme.")
        d.append(d1)
        d.append(GXBitString(self.__rawValue, 0))
        return GXAsn1Converter.toByteArray(d)

    def toPem(self):
        """
        Get public key as PEM format.
        """
        return (
            "-----BEGIN PUBLIC KEY-----\n"
            + self.toDer()
            + "\n-----END PUBLIC KEY-----\n"
        )

    def __str__(self):
        if self.scheme == Ecc.P256:
            sb = "NIST P-256"
            size = 32
        elif self.scheme == Ecc.P384:
            sb = "NIST P-384"
            size = 48
        else:
            raise ValueError("Invalid scheme.")
        sb += " public x coord: 0x"
        sb += format(
            int.from_bytes(self.__rawValue[1 : 1 + size], byteorder="big"), "X"
        )
        sb += "\r public y coord: 0x"
        sb += (
            format(int.from_bytes(self.__rawValue[1 + size :], byteorder="big"), "X")
            + "\r"
        )
        return sb

    def __eq__(self, other):
        return isinstance(other, GXPublicKey) and self.__rawValue == other.rawValue



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http:#www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from gurux_dlms.asn.enums.X9ObjectIdentifier import X9ObjectIdentifier


class X9ObjectIdentifierConverter:
    __OID_STRING_MAP = {
        X9ObjectIdentifier.ID_FIELD_TYPE: "1.2.840.10045.1",
        X9ObjectIdentifier.PRIME_FIELD: "1.2.840.10045.1",
        X9ObjectIdentifier.CHARACTERISTIC_TWO_FIELD: "1.2.840.10045.1.2",
        X9ObjectIdentifier.GN_BASIS: "1.2.840.10045.1.2.3.1",
        X9ObjectIdentifier.TP_BASIS: "1.2.840.10045.1.2.3.2",
        X9ObjectIdentifier.PP_BASIS: "1.2.840.10045.1.2.3.3",
        X9ObjectIdentifier.EC_DSA_WITH_SHA1: "1.2.840.10045.4.1",
        X9ObjectIdentifier.ID_EC_PUBLIC_KEY: "1.2.840.10045.2.1",
        X9ObjectIdentifier.EC_DSA_WITH_SHA2: "1.2.840.10045.4.3",
        X9ObjectIdentifier.EC_DSA_WITH_SHA224: "1.2.840.10045.4.31",
        X9ObjectIdentifier.EC_DSA_WITH_SHA256: "1.2.840.10045.4.32",
        X9ObjectIdentifier.EC_DSA_WITH_SHA384: "1.2.840.10045.4.33",
        X9ObjectIdentifier.EC_DSA_WITH_SHA512: "1.2.840.10045.4.34",
        X9ObjectIdentifier.ELLIPTIC_CURVE: "1.2.840.10045.3",
        X9ObjectIdentifier.C_TWO_CURVE: "1.2.840.10045.3.0",
        X9ObjectIdentifier.C2PNB163V1: "1.2.840.10045.3.0.1",
        X9ObjectIdentifier.C2PNB163V2: "1.2.840.10045.3.0.2",
        X9ObjectIdentifier.C2PNB163V3: "1.2.840.10045.3.0.3",
        X9ObjectIdentifier.C2PNB176W1: "1.2.840.10045.3.0.4",
        X9ObjectIdentifier.C2TNB191V1: "1.2.840.10045.3.0.5",
        X9ObjectIdentifier.C2TNB191V2: "1.2.840.10045.3.0.6",
        X9ObjectIdentifier.C2TNB191V3: "1.2.840.10045.3.0.7",
        X9ObjectIdentifier.C2ONB191V4: "1.2.840.10045.3.0.8",
        X9ObjectIdentifier.C2ONB191V5: "1.2.840.10045.3.0.9",
        X9ObjectIdentifier.C2PNB208W1: "1.2.840.10045.3.0.10",
        X9ObjectIdentifier.C2TNB239V1: "1.2.840.10045.3.0.11",
        X9ObjectIdentifier.C2TNB239V2: "1.2.840.10045.3.0.12",
        X9ObjectIdentifier.C2TNB239V3: "1.2.840.10045.3.0.13",
        X9ObjectIdentifier.C2ONB239V4: "1.2.840.10045.3.0.14",
        X9ObjectIdentifier.C2ONB239V5: "1.2.840.10045.3.0.15",
        X9ObjectIdentifier.C2PNB272W1: "1.2.840.10045.3.0.16",
        X9ObjectIdentifier.C2PNB304W1: "1.2.840.10045.3.0.17",
        X9ObjectIdentifier.C2TNB359V1: "1.2.840.10045.3.0.18",
        X9ObjectIdentifier.C2PNB368W1: "1.2.840.10045.3.0.19",
        X9ObjectIdentifier.C2TNB431R1: "1.2.840.10045.3.0.20",
        X9ObjectIdentifier.PRIME_CURVE: "1.2.840.10045.3.1",
        X9ObjectIdentifier.PRIME192V1: "1.2.840.10045.3.1.1",
        X9ObjectIdentifier.PRIME192V2: "1.2.840.10045.3.1.2",
        X9ObjectIdentifier.PRIME192V3: "1.2.840.10045.3.1.3",
        X9ObjectIdentifier.PRIME239V1: "1.2.840.10045.3.1.4",
        X9ObjectIdentifier.PRIME239V2: "1.2.840.10045.3.1.5",
        X9ObjectIdentifier.PRIME239V3: "1.2.840.10045.3.1.6",
        X9ObjectIdentifier.PRIME256V1: "1.2.840.10045.3.1.7",
        X9ObjectIdentifier.ID_DSA: "1.2.840.10040.4.1",
        X9ObjectIdentifier.ID_DSA_WITH_SHA1: "1.2.840.10040.4.3",
        X9ObjectIdentifier.X9X63SCHEME: "1.3.133.16.840.63.0",
        X9ObjectIdentifier.DH_SINGLE_PASS_STD_DH_SHA1KDF_SCHEME: "1.3.133.16.840.63.0.2",
        X9ObjectIdentifier.DH_SINGLE_PASS_COFACTOR_DH_SHA1KDF_SCHEME: "1.3.133.16.840.63.0.3",
        X9ObjectIdentifier.MQV_SINGLE_PASS_SHA1KDF_SCHEME: "1.3.133.16.840.63.0.16",
        X9ObjectIdentifier.ANSI_X9_42: "1.2.840.10046",
        X9ObjectIdentifier.DH_PUBLIC_NUMBER: "1.2.840.10046.2.1",
        X9ObjectIdentifier.X9X42SCHEMES: "1.2.840.10046.2.3",
        X9ObjectIdentifier.DH_STATIC: "1.2.840.10046.2.3.1",
        X9ObjectIdentifier.DH_EPHEM: "1.2.840.10046.2.3.2",
        X9ObjectIdentifier.DH_ONE_FLOW: "1.2.840.10046.2.3.3",
        X9ObjectIdentifier.DH_HYBRID1: "1.2.840.10046.2.3.4",
        X9ObjectIdentifier.DH_HYBRID2: "1.2.840.10046.2.3.5",
        X9ObjectIdentifier.DH_HYBRID_ONE_FLOW: "1.2.840.10046.2.3.6",
        X9ObjectIdentifier.MQV2: "1.2.840.10046.2.3.7",
        X9ObjectIdentifier.MQV1: "1.2.840.10046.2.3.8",
        X9ObjectIdentifier.SECP384R1: "1.3.132.0.34",
    }

    @classmethod
    def getString(cls, value):
        try:
            return cls.__OID_STRING_MAP[value]
        except KeyError:
            raise KeyError(f"Invalid X509Name. {value}")

    @classmethod
    def fromString(cls, value):
        for enum_value, oid in cls.__OID_STRING_MAP.items():
            if oid == value:
                return enum_value
        raise KeyError(f"Unknown OID string: {value}")



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http:#www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
# pylint: disable=too-many-instance-attributes
class GXDLMSRoutingConfiguration:
    """
    The routing configuration element specifies all parameters linked to the routing mechanism described in ITU-T G.9903:2014.
    """

    def __init__(self):
        """
        Constructor.
        """

        self.__netTraversalTime = 0
        self.__routingTableEntryTtl = 0
        self.__kr = 0
        self.__km = 0
        self.__kc = 0
        self.__kq = 0
        self.__kh = 0
        self.__krt = 0
        self.__rreqRetries = 0
        self.__rreqReqWait = 0
        self.__blacklistTableEntryTtl = 0
        self.__unicastRreqGenEnable = False
        self.__rlcEnabled = False
        self.__addRevLinkCost = 0

    @property
    def netTraversalTime(self):
        """
        Maximum time that a packet is expected to take to reach any node from any node in seconds.
        """
        return self.__netTraversalTime

    @netTraversalTime.setter
    def netTraversalTime(self, value):
        self.__netTraversalTime = value

    @property
    def routingTableEntryTtl(self):
        """
        Maximum time-to-live of a routing table entry (in minutes).
        """
        return self.__routingTableEntryTtl

    @routingTableEntryTtl.setter
    def routingTableEntryTtl(self, value):
        self.__routingTableEntryTtl = value

    @property
    def kr(self):
        """
        A weight factor for the Robust Mode to calculate link cost.
        """
        return self.__kr

    @kr.setter
    def kr(self, value):
        self.__kr = value

    @property
    def km(self):
        """
        A weight factor for modulation to calculate link cost.
        """
        return self.__km

    @km.setter
    def km(self, value):
        self.__km = value

    @property
    def kc(self):
        """
        A weight factor for number of active tones to calculate link cost.
        """
        return self.__kc

    @kc.setter
    def kc(self, value):
        self.__kc = value

    @property
    def kq(self):
        """
        A weight factor for LQI to calculate route cost.
        """
        return self.__kq

    @kq.setter
    def kq(self, value):
        self.__kq = value

    @property
    def kh(self):
        """
        A weight factor for hop to calculate link cost.
        """
        return self.__kh

    @kh.setter
    def kh(self, value):
        self.__kh = value

    @property
    def krt(self):
        """
        A weight factor for the number of active routes in the routing table to calculate link cost.
        """
        return self.__krt

    @krt.setter
    def krt(self, value):
        self.__krt = value

    @property
    def rreqRetries(self):
        """
        The number of RREQ retransmission in case of RREP reception time out.
        """
        return self.__rreqRetries

    @rreqRetries.setter
    def rreqRetries(self, value):
        self.__rreqRetries = value

    @property
    def rreqReqWait(self):
        """
        The number of seconds to wait between two consecutive RREQ � RERR generations.
        """
        return self.__rreqReqWait

    @rreqReqWait.setter
    def rreqReqWait(self, value):
        self.__rreqReqWait = value

    @property
    def blacklistTableEntryTtl(self):
        """
        Maximum time-to-live of a blacklisted neighbour entry (in minutes).
        """
        return self.__blacklistTableEntryTtl

    @blacklistTableEntryTtl.setter
    def blacklistTableEntryTtl(self, value):
        self.__blacklistTableEntryTtl = value

    @property
    def unicastRreqGenEnable(self):
        """
        If TRUE, the RREQ shall be generated with its 'unicast RREQ'.
        """
        return self.__unicastRreqGenEnable

    @unicastRreqGenEnable.setter
    def unicastRreqGenEnable(self, value):
        self.__unicastRreqGenEnable = value

    @property
    def rlcEnabled(self):
        """
        Enable the sending of RLCREQ frame by the device.
        """
        return self.__rlcEnabled

    @rlcEnabled.setter
    def rlcEnabled(self, value):
        self.__rlcEnabled = value

    @property
    def addRevLinkCost(self):
        """
        It represents an additional cost to take into account a possible asymmetry in the link.
        """
        return self.__addRevLinkCost

    @addRevLinkCost.setter
    def addRevLinkCost(self, value):
        self.__addRevLinkCost = value



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http:#www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http:#www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXEccPoint import GXEccPoint


class _GXShamirs:
    """
    This class implements GXShamir's trick.
    """

    @classmethod
    def __pointAdd(cls, curve, ret, p1, p2):
        """
        Add points.

            Parameters:
                curve: Used curve.
                ret: Result.
                p1: Point 1.
                p2: Point 2.
        """
        # Calculate lambda.
        ydiff = p2.y
        ydiff -= p1.y
        xdiff = p2.x
        xdiff -= p1.x
        xdiff = pow(xdiff, -1, curve.p)
        lambda_ = ydiff
        lambda_ *= xdiff
        lambda_ %= curve.p
        # calculate resulting x coord.
        ret.x = lambda_
        ret.x *= lambda_
        ret.x -= p1.x
        ret.x -= p2.x
        ret.x %= curve.p
        # calculate resulting y coord
        ret.y = p1.x
        ret.y -= ret.x
        ret.y *= lambda_
        ret.y -= p1.y
        ret.y %= curve.p

    @classmethod
    def _pointDouble(cls, curve, ret, p1):
        """
        Double point.

            Parameters:
                curve: Used curve.
                ret: Result value.
                p1: Point to double
        """
        numer = p1.x
        numer *= p1.x
        numer *= 3
        numer += curve.a
        denom = p1.y
        denom *= 2
        denom = pow(denom, -1, curve.p)
        lambda_ = int(numer)
        lambda_ *= denom
        lambda_ %= curve.p
        # calculate resulting x coord
        ret.x = lambda_
        ret.x *= lambda_
        ret.x -= p1.x
        ret.x -= p1.x
        ret.x %= curve.p
        # calculate resulting y coord
        ret.y = p1.x
        ret.y -= ret.x
        ret.y *= lambda_
        ret.y -= p1.y
        ret.y %= curve.p

    @classmethod
    def pointMulti(cls, curve, ret, point, scalar):
        """
        Multiply point with big integer value.

            Parameters:
                curve: Used curve.
                ret: Return value.
                point: Point.
                scalar: Scaler.
        """
        R0 = GXEccPoint(point.x, point.y)
        R1 = GXEccPoint()
        tmp = GXEccPoint()
        cls._pointDouble(curve, R1, point)
        dbits = scalar.bit_length()
        dbits -= 2
        while True:
            if scalar & 1 << dbits != 0:
                # If bit is set.
                tmp.x = R0.x
                tmp.y = R0.y
                cls.__pointAdd(curve, R0, R1, tmp)
                tmp.x = R1.x
                tmp.y = R1.y
                cls._pointDouble(curve, R1, tmp)
            else:
                tmp.x = R1.x
                tmp.y = R1.y
                cls.__pointAdd(curve, R1, R0, tmp)
                tmp.x = R0.x
                tmp.y = R0.y
                cls._pointDouble(curve, R0, tmp)
            if dbits == 0:
                break
            dbits -= 1
        ret.x = R0.x
        ret.y = R0.y

    # pylint: disable=too-many-arguments
    @classmethod
    def trick(cls, curve, pub, ret, u1, u2):
        """
        Count Shamir's trick.

            Parameters:
                curve: Used curve.
                pub: Public key.
                ret: Result.
                u1:
                u2:
        """
        sum_ = GXEccPoint()
        op2 = GXEccPoint(
            int.from_bytes(pub.x, byteorder="big"),
            int.from_bytes(pub.y, byteorder="big"),
        )
        cls.__pointAdd(curve, sum_, curve.g, op2)
        bits1 = u1.bit_length()
        bits2 = u2.bit_length()
        if bits1 > bits2:
            pos = bits1
        else:
            pos = bits2
        pos -= 1
        if u1 & 1 << pos != 0 and u2 & 1 << pos != 0:
            ret.x = sum_.x
            ret.y = sum_.y
        elif u1 & 1 << pos != 0:
            ret.x = curve.g.x
            ret.y = curve.g.y
        elif u2 & 1 << pos != 0:
            ret.x = op2.x
            ret.y = op2.y
        tmp = GXEccPoint()
        pos -= 1
        while True:
            tmp.x = ret.x
            tmp.y = ret.y
            cls._pointDouble(curve, ret, tmp)
            tmp.x = ret.x
            tmp.y = ret.y
            if u1 & 1 << pos != 0 and u2 & 1 << pos != 0:
                cls.__pointAdd(curve, ret, tmp, sum_)
            elif u1 & 1 << pos != 0:
                cls.__pointAdd(curve, ret, tmp, curve.g)
            elif u2 & 1 << pos != 0:
                cls.__pointAdd(curve, ret, tmp, op2)
            if pos == 0:
                break
            pos -= 1



#
#  --------------------------------------------------------------------------
#   Gurux Ltd
#
#
#
#  Filename: $HeadURL$
#
#  Version: $Revision$,
#                   $Date$
#                   $Author$
#
#  Copyright (c) Gurux Ltd
#
# ---------------------------------------------------------------------------
#
#   DESCRIPTION
#
#  This file is a part of Gurux Device Framework.
#
#  Gurux Device Framework is Open Source software; you can redistribute it
#  and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; version 2 of the License.
#  Gurux Device Framework is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#  See the GNU General Public License for more details.
#
#  More information of Gurux products: http://www.gurux.org
#
#  This code is licensed under the GNU General Public License v2.
#  Full text may be retrieved at http://www.gnu.org/licenses/gpl-2.0.txt
# ---------------------------------------------------------------------------
from .GXDLMSObject import GXDLMSObject
from .IGXDLMSBase import IGXDLMSBase
from ..enums import ErrorCode
from ..internal._GXCommon import _GXCommon
from ..enums import ObjectType, DataType
from ..internal._GXLocalizer import _GXLocalizer

# pylint: disable=too-many-instance-attributes
class GXDLMSLlcSscsSetup(GXDLMSObject, IGXDLMSBase):
    """
    Online help:
    http://www.gurux.fi/Gurux.DLMS.Objects.GXDLMSLlcSscsSetup
    """

    def __init__(self, ln="0.0.28.0.0.255", sn=0):
        """
        Constructor.

        ln : Logical Name of the object.
        sn : Short Name of the object.
        """
        GXDLMSObject.__init__(self, ObjectType.LLC_SSCS_SETUP, ln, sn)
        self.serviceNodeAddress = 0
        self.baseNodeAddress = 0

    def getValues(self):
        return [self.logicalName,
                self.serviceNodeAddress,
                self.baseNodeAddress]

    def reset(self, client):
        """Reset value."""
        return client.method(self, 1, 0, DataType.INT8)

    #
    # Returns collection of attributes to read.  If attribute is static and
    # already read or device is returned HW error it is not returned.
    #
    def getAttributeIndexToRead(self, all_):
        attributes = []
        #  LN is static and read only once.
        if all_ or not self.logicalName:
            attributes.append(1)
        #  ServiceNodeAddress
        if all_ or self.canRead(2):
            attributes.append(2)
        #  BaseNodeAddress
        if all_ or self.canRead(3):
            attributes.append(3)
        return attributes

    def invoke(self, settings, e):
        #  Resets the value to the default value.
        #  The default value is an instance specific constant.
        if e.index == 1:
            self.serviceNodeAddress = 0xFFE
            self.baseNodeAddress = 0
        else:
            e.error = ErrorCode.READ_WRITE_DENIED
    #
    # Returns amount of attributes.
    #
    def getAttributeCount(self):
        return 3

    #
    # Returns amount of methods.
    #
    def getMethodCount(self):
        return 1

    def getNames(self):
        return (_GXLocalizer.gettext("Logical name"),\
            _GXLocalizer.gettext("Service node address"),\
            _GXLocalizer.gettext("Base node address"))

    def getMethodNames(self):
        return (_GXLocalizer.gettext("Reset"),)

    def getDataType(self, index):
        if index == 1:
            return DataType.OCTET_STRING
        if index in (2, 3):
            return DataType.UINT16
        raise ValueError("getDataType failed. Invalid attribute index.")

    #
    # Returns value of given attribute.
    #
    def getValue(self, settings, e):
        if e.index == 1:
            return _GXCommon.logicalNameToBytes(self.logicalName)
        if e.index == 2:
            return self.serviceNodeAddress
        if e.index == 3:
            return self.baseNodeAddress
        e.error = ErrorCode.READ_WRITE_DENIED
        return None

    #
    # Set value of given attribute.
    #
    def setValue(self, settings, e):
        if e.index == 1:
            self.logicalName = _GXCommon.toLogicalName(e.value)
        elif e.index == 2:
            self.serviceNodeAddress = e.value
        elif e.index == 3:
            self.baseNodeAddress = e.value
        else:
            e.error = ErrorCode.READ_WRITE_DENIED

    def load(self, reader):
        self.serviceNodeAddress = reader.readElementContentAsInt("ServiceNodeAddress")
        self.baseNodeAddress = reader.readElementContentAsInt("BaseNodeAddress")

    def save(self, writer):
        writer.writeElementString("ServiceNodeAddress", self.serviceNodeAddress)
        writer.writeElementString("BaseNodeAddress", self.baseNodeAddress)
